/* //device/system/reference-ril/reference-ril.c
**
** Copyright 2006, The Android Open Source Project
**
** Licensed under the Apache License, Version 2.0 (the "License");
** you may not use this file except in compliance with the License.
** You may obtain a copy of the License at
**
**     http://www.apache.org/licenses/LICENSE-2.0
**
** Unless required by applicable law or agreed to in writing, software
** distributed under the License is distributed on an "AS IS" BASIS,
** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
** See the License for the specific language governing permissions and
** limitations under the License.
*/

#include <telephony/ril_cdma_sms.h>
#include <telephony/librilutils.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <alloca.h>
#include "atchannel.h"
#include "at_tok.h"
#include "fcp_parser.h"
#include "misc.h"
#include <getopt.h>
#include <sys/socket.h>
#include <cutils/sockets.h>
#include <cutils/properties.h>
#include <termios.h>
#include <dirent.h>
#include <net/if.h>
#include <net/route.h>
#include <arpa/inet.h>
//#include <sys/system_properties.h>
#include "ril.h"
//#include "hardware/qemu_pipe.h"

#define LOG_TAG "RIL"
#include <utils/Log.h>

//static void *noopRemoveWarning( void *a ) { return a; }
//#define RIL_UNUSED_PARM(a) noopRemoveWarning((void *)&(a));

//#define MAX_AT_RESPONSE 0x1000

/* pathname returned from RIL_REQUEST_SETUP_DATA_CALL / RIL_REQUEST_SETUP_DEFAULT_PDP */
//#define PPP_TTY_PATH 	"/dev/ttyUSB1"

// Default MTU value
#define DEFAULT_MTU 1500

char PORT_ECM_KEYNAME[20] = "";

/* Modem Technology bits */
#define MDM_UNKNOWN       	0x00
#define MDM_GSM         	0x01
#define MDM_WCDMA       	0x02
#define MDM_CDMA        	0x04
#define MDM_EVDO        	0x08
#define MDM_TD_SCDMA    	0x10
#define MDM_LTE         	0x20

typedef struct {
    int supportedTechs; // Bitmask of supported Modem Technology bits
    int currentTech;    // Technology the modem is currently using (in the format used by modem)
	//int isMultimode;

    // Preferred mode bitmask. This is actually 4 byte-sized bitmasks with different priority values,
    // in which the byte number from LSB to MSB give the priority.
    //
    //          |MSB|   |   |LSB
    // value:   |00 |00 |00 |00
    // byte #:  |3  |2  |1  |0
    //
    // Higher byte order give higher priority. Thus, a value of 0x0000000f represents
    // a preferred mode of GSM, WCDMA, CDMA, and EvDo in which all are equally preferrable, whereas
    // 0x00000201 represents a mode with GSM and WCDMA, in which WCDMA is preferred over GSM
    int32_t preferredNetworkMode;
    int subscription_source;
    int usePPP;
    int isPppStart;
	int useVoLTE;
	int callsNum;
} ModemInfo;

static ModemInfo *sMdmInfo;
// TECH returns the current technology in the format used by the modem.
// It can be used as an l-value
#define TECH(mdminfo)                 ((mdminfo)->currentTech)
// TECH_BIT returns the bitmask equivalent of the current tech
#define TECH_BIT(mdminfo)            (1 << ((mdminfo)->currentTech))
//#define IS_MULTIMODE(mdminfo)         ((mdminfo)->isMultimode)
#define TECH_SUPPORTED(mdminfo, tech) ((mdminfo)->supportedTechs & (tech))
#define PREFERRED_NETWORK(mdminfo)    ((mdminfo)->preferredNetworkMode)
// CDMA Subscription Source
#define SSOURCE(mdminfo)              ((mdminfo)->subscription_source)
// use ppp
#define USE_PPP(mdminfo)				((mdminfo)->usePPP)
#define USE_VOLTE(mdminfo)				((mdminfo)->useVoLTE)
#define PPPD_START(mdminfo)				((mdminfo)->isPppStart)
#define CALLS_NUM(mdminfo)				((mdminfo)->callsNum)

/* Last call fail cause */
static int cc_lastCallFailCause = CALL_FAIL_ERROR_UNSPECIFIED;
static int s_lastPdpFailCause = PDP_FAIL_ERROR_UNSPECIFIED;

pthread_t s_tid_mainloop;

#define PPP_PDP_TYPE_LEN 10
#define ADDRESS_LEN 80

typedef struct{
	char pdptype[PPP_PDP_TYPE_LEN+1];
	char ip_addr[ADDRESS_LEN+1];
	char dns_addr[2*ADDRESS_LEN+1];
	char gateways[ADDRESS_LEN+1];
	char pcscf_addr[2*ADDRESS_LEN+1];
}Rmnet_Parameters;

Rmnet_Parameters rmnetParam;

typedef enum {
    SIM_ABSENT = 0,
    SIM_NOT_READY = 1,
    SIM_READY = 2, /* SIM_READY means the radio state is RADIO_STATE_SIM_READY */
    SIM_PIN = 3,
    SIM_PUK = 4,
    SIM_NETWORK_PERSONALIZATION = 5,
    SIM_PIN2 = 6,
    SIM_PUK2 = 7,
} SIM_Status;

typedef enum{
	PPP_CLOSING = 0,
	PPP_CLOSED = 1,
	PPP_STARTING = 2,
	PPP_STARTED = 3,
}PPP_STATUS;

typedef enum {
    UICC_TYPE_UNKNOWN,
    UICC_TYPE_SIM,
    UICC_TYPE_USIM,
    UICC_TYPE_UIM,
} UICC_Type;

#define OUTSTANDING_CMT_SMS 	0
#define OUTSTANDING_CDS_REPORT	1

typedef struct wait_pdu_s
{
    int type;
    char* pdu;
    struct wait_pdu_s* next;
} wait_pdu;

wait_pdu* g_wait_pdu = NULL;
int g_outstanding_sms = 0;
static pthread_mutex_t s_pdu_mutex = PTHREAD_MUTEX_INITIALIZER;

static void onRequest (int request, void *data, size_t datalen, RIL_Token t);
static RIL_RadioState getRadioState();
static int onSupports (int requestCode);
static void onCancel (RIL_Token t);
static const char *getVersion();
static int isRadioOn();
static SIM_Status getSIMStatus();
static int getCardStatus(RIL_CardStatus_v6 **pp_card_status);
static void freeCardStatus(RIL_CardStatus_v6 *p_card_status);
static void onDataCallListChanged(void *param);
static UICC_Type getUICCType(void);
static void onDeactiveDataCallList(void);
extern const char * requestToString(int request);
static int getAddressInfo();
static int getAddrFromModem(int cid);
static void getIccid(void);


/*** Static Variables ***/
static const RIL_RadioFunctions s_callbacks = {
    RIL_VERSION,
    onRequest,
    getRadioState,
    onSupports,
    onCancel,
    getVersion
};

#ifdef RIL_SHLIB
static const struct RIL_Env *s_rilenv;
#define RIL_onRequestComplete(t, e, response, responselen) s_rilenv->OnRequestComplete(t,e, response, responselen)
#define RIL_onUnsolicitedResponse(a,b,c) s_rilenv->OnUnsolicitedResponse(a,b,c)
#define RIL_requestTimedCallback(a,b,c) s_rilenv->RequestTimedCallback(a,b,c)
#endif

static RIL_RadioState sState = RADIO_STATE_UNAVAILABLE;

static pthread_mutex_t s_state_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t s_state_cond = PTHREAD_COND_INITIALIZER;

static int s_port = -1;
static const char * s_device_path = NULL;
static int          s_device_socket = 0;

/* trigger change to this with s_state_cond */
static int s_closed = 0;

static const struct timeval TIMEVAL_SIMPOLL = {1,0};
static const struct timeval TIMEVAL_CALLSTATEPOLL = {0,500000};
static const struct timeval TIMEVAL_0 = {0,0};


static int s_cell_info_rate_ms = INT_MAX;
static int s_mcc = 0;
static int s_mnc = 0;
static int s_lac = 0;
static int s_cid = 0;

static void pollSIMState (void *param);
static void setRadioState(RIL_RadioState newState);


static int techFromModemType(int mdmtype);

static int clccStateToRILState(int state, RIL_CallState *p_state)

{
    switch(state) {
        case 0: *p_state = RIL_CALL_ACTIVE;   return 0;
        case 1: *p_state = RIL_CALL_HOLDING;  return 0;
        case 2: *p_state = RIL_CALL_DIALING;  return 0;
        case 3: *p_state = RIL_CALL_ALERTING; return 0;
        case 4: *p_state = RIL_CALL_INCOMING; return 0;
        case 5: *p_state = RIL_CALL_WAITING;  return 0;
        default: return -1;
    }
}


static int cmiot_setprop(char *property, char *value)
{
	char * cmdSet=NULL;
	int err;
	if(NULL != value){
	   asprintf(&cmdSet, "/system/bin/setprop %s %s", property,value); 
	}else{
		asprintf(&cmdSet, "/system/bin/setprop %s \"\"", property);
	}

	RLOGD("setprop comand is: %s", cmdSet);
	err = system(cmdSet);
	free(cmdSet);
	return err;
}

/** do post-AT+CFUN=1 initialization */
static void onRadioPowerOn()
{
   RLOGI("onRadioPowerOn");
    pollSIMState(NULL);
}

/** do post- SIM ready initialization */
static void onSIMReady()
{   
    at_send_command_singleline("AT+CSMS=1", "+CSMS:", NULL);
    /*
     * Always send SMS messages directly to the TE
     *
     * mode = 1 // discard when link is reserved (link should never be
     *             reserved)
     * mt = 2   // most messages routed to TE
     * bm = 2   // new cell BM's routed to TE
     * ds = 1   // Status reports routed to TE
     * bfr = 1  // flush buffer
     */
     
	
	
     /* Always send SMS messages directly to the TE */
     at_send_command("AT+CNMI=2,1,2,2,0", NULL);
     if (isRadioOn()!=1) {
		at_send_command("AT+CFUN=1",NULL);
	
     }
}

static void requestRadioPower(void *data, size_t datalen, RIL_Token t)
{
    int onOff;

    int err;


	
    ATResponse *p_response = NULL;

   RLOGI("requestRadioPower()");
	
    assert (datalen >= sizeof(int *));
    onOff = ((int *)data)[0];

   RLOGI("requestRadioPower onOff=%d",onOff);
   
    if (onOff == 0 && getRadioState() != RADIO_STATE_OFF) 
	{
		err = at_send_command("AT+CFUN=0", &p_response);
		if (err < 0 || p_response->success == 0) goto error;
        setRadioState(RADIO_STATE_OFF);
    } else if (onOff > 0 && getRadioState() == RADIO_STATE_OFF) {
        err = at_send_command("AT+CFUN=1", &p_response);
        if (err < 0|| p_response->success == 0) {
            // Some stacks return an error when there is no SIM,
            // but they really turn the RF portion on
            // So, if we get an error, let's check to see if it
            // turned on anyway

            if (isRadioOn() != 1) {
                goto error;
            }
        }
		at_send_command("AT+CGATT=1",NULL);	//if current network mode is 2G, needed this
        setRadioState(RADIO_STATE_SIM_NOT_READY);
    }

    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
error:
    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestShutdown(RIL_Token t)
{
    int onOff;

    int err;
    ATResponse *p_response = NULL;
   RLOGI("requestShutdown()");

    if (getRadioState() != RADIO_STATE_OFF) {
        err = at_send_command("AT+CFUN=0", &p_response);
        setRadioState(RADIO_STATE_UNAVAILABLE);
    }

    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;
}

static void requestOrSendDataCallList(RIL_Token *t);

static void onDataCallListChanged(void *param __unused)
{
    requestOrSendDataCallList(NULL);
}

static void requestDataCallList(void *data __unused, size_t datalen __unused, RIL_Token t)
{
    requestOrSendDataCallList(&t);
}

static void  requestOrSendDataCallList(RIL_Token *t)
{
    ATResponse *p_response;
    ATLine *p_cur;
    int err;
    int n = 0;
    char *line,*out;
	int active,cid;

    err = at_send_command_multiline ("AT+CGACT?", "+CGACT:", &p_response);
    if (err < 0 || p_response->success == 0) 
        goto error;


    for (p_cur = p_response->p_intermediates; p_cur != NULL;
         p_cur = p_cur->p_next)
        n++;

	RLOGD("requestOrSendDataCallList:n=%d",n);
	n = 1;
    RIL_Data_Call_Response_v6 *response = alloca(n * sizeof(RIL_Data_Call_Response_v6));

    response->status = -1;
    response->suggestedRetryTime = -1;
    response->cid = -1;
    response->active = -1;
    response->type = (char*)"";
	response->ifname = PORT_ECM_KEYNAME;
    response->addresses = (char*)"";
    response->dnses = (char*)"";
    response->gateways = (char*)"";

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &response->cid);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &response->active);
    if (err < 0) goto error;

	if(PPPD_START(sMdmInfo) == PPP_STARTED && response->active == 1)			
		response->active = 2;

    at_response_free(p_response);

    err = at_send_command_multiline ("AT+CGDCONT?", "+CGDCONT:", &p_response);
    if (err < 0 || p_response->success == 0)
        goto error;

    line = p_response->p_intermediates->line;
	
    err = at_tok_start(&line);
	if (err < 0) goto error;

    err = at_tok_nextint(&line, &cid);
    if (err < 0) goto error;

    // Assume no error
    response->status = 0;

    // type
    err = at_tok_nextstr(&line, &out);
    if (err < 0) goto error;
    response->type = (char*)alloca(strlen(out) + 1);
    strcpy(response->type, out);
        
	response->ifname = PORT_ECM_KEYNAME;
	
	if(USE_PPP(sMdmInfo))
		err = getAddressInfo();
	else
		err = getAddrFromModem(cid);
	
	if(err == 0){
		response->addresses = rmnetParam.ip_addr;
		response->dnses = rmnetParam.dns_addr;
		response->gateways = rmnetParam.gateways;
	}
		

    if (t != NULL)
        RIL_onRequestComplete(*t, RIL_E_SUCCESS, response,n * sizeof(RIL_Data_Call_Response_v6));
    else
        RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED, response,n * sizeof(RIL_Data_Call_Response_v6));

	at_response_free(p_response);
    return;

error:
    if (t != NULL)
        RIL_onRequestComplete(*t, RIL_E_GENERIC_FAILURE, NULL, 0);
    else
        RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED, NULL, 0);

    at_response_free(p_response);
}

static void requestQueryNetworkSelectionMode(void *data __unused, size_t datalen __unused, RIL_Token t)
{
    int err;
    ATResponse *p_response = NULL;
    int response = 0;
    char *line;

    err = at_send_command_singleline("AT+COPS?", "+COPS:", &p_response);

    if (err < 0 || p_response->success == 0) {
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);

    if (err < 0) {
        goto error;
    }

    err = at_tok_nextint(&line, &response);

    if (err < 0) {
        goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(int));
    at_response_free(p_response);
    return;

error:
    at_response_free(p_response);
    RLOGE("requestQueryNetworkSelectionMode must never return error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}


void requestSetNetworkSelectionManual(void *data, size_t datalen,RIL_Token t)
{
    /*
     * AT+COPS=[<mode>[,<format>[,<oper>[,<AcT>]]]]
     *    <mode>   = 4 = Manual (<oper> field shall be present and AcT optionally) with fallback to automatic if manual fails.
     *    <format> = 2 = Numeric <oper>, the number has structure:
     *                   (country code digit 3)(country code digit 2)(country code digit 1)
     *                   (network code digit 2)(network code digit 1)
     */

    (void) datalen;
    int err = 0;
    const char *mccMnc = (const char *) data;
	char *cmd;

    /* Check inparameter. */
    if (mccMnc == NULL)
        goto error;

	asprintf(&cmd,"AT+COPS=1,2,\"%s\"",mccMnc);
    /* Build and send command. */
    err = at_send_command(cmd,NULL);
	free(cmd);
    if (err != 0) {
        err = at_send_command("AT+COPS=0",NULL);
        if(err != 0)
            goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}


static void sendCallStateChanged(void *param __unused)
{
    RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_CALL_STATE_CHANGED,NULL, 0);
}

/**
 * Note: directly modified line and has *p_call point directly into
 * modified line
 */
static int callFromCLCCLine(char *line, RIL_Call *p_call)
{
    int err;
    int state;
    int mode;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(p_call->index));
    if (err < 0) goto error;

    err = at_tok_nextbool(&line, &(p_call->isMT));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &state);
    if (err < 0) goto error;

    err = clccStateToRILState(state, &(p_call->state));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &mode);
    if (err < 0) goto error;

    p_call->isVoice = (mode == 0);

    err = at_tok_nextbool(&line, &(p_call->isMpty));
    if (err < 0) goto error;

    if (at_tok_hasmore(&line)) {
        err = at_tok_nextstr(&line, &(p_call->number));

        /* tolerate null here */
        if (err < 0) return 0;

        // Some lame implementations return strings
        // like "NOT AVAILABLE" in the CLCC line
        if (p_call->number != NULL
            && 0 == strspn(p_call->number, "+0123456789")
        ) {
            p_call->number = NULL;
        }

        err = at_tok_nextint(&line, &p_call->toa);
        if (err < 0) goto error;
    }

    p_call->uusInfo = NULL;

    return 0;

error:
    RLOGE("invalid CLCC line\n");
    return -1;
}


static void requestGetCurrentCalls(void *data __unused, size_t datalen __unused, RIL_Token t)
{
    int err;
    ATResponse *p_response;
    ATLine *p_cur;
    int countCalls;
    int countValidCalls;
    RIL_Call *p_calls;
    RIL_Call **pp_calls;
    int i;
    int needRepoll = 0;

    err = at_send_command_multiline ("AT+CLCC", "+CLCC:", &p_response);

    if (err != 0 || p_response->success == 0)
		goto error;
    
    /* count the calls */
    for (countCalls = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next
    ) {
        countCalls++;
    }

	CALLS_NUM(sMdmInfo) = countCalls;

    /* yes, there's an array of pointers and then an array of structures */

    pp_calls = (RIL_Call **)alloca(countCalls * sizeof(RIL_Call *));
    p_calls = (RIL_Call *)alloca(countCalls * sizeof(RIL_Call));
    memset (p_calls, 0, countCalls * sizeof(RIL_Call));

    /* init the pointer array */
    for(i = 0; i < countCalls ; i++) {
        pp_calls[i] = &(p_calls[i]);
    }

    for (countValidCalls = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next
    ) {
        err = callFromCLCCLine(p_cur->line, p_calls + countValidCalls);

        if (err != 0) {
            continue;
        }


        if (p_calls[countValidCalls].state != RIL_CALL_ACTIVE
            && p_calls[countValidCalls].state != RIL_CALL_HOLDING
        ) {
            needRepoll = 1;
        }
        if(p_calls[countValidCalls].isVoice) // only count voice calls
        	countValidCalls++;
    }

	RLOGI("Calls=%d,Valid=%d\n",countCalls,countValidCalls);

    RIL_onRequestComplete(t, RIL_E_SUCCESS, pp_calls,countValidCalls * sizeof (RIL_Call *));

    at_response_free(p_response);

	// We don't seem to get a "NO CARRIER" message from
                            // smd, so we're forced to poll until the call ends.
	if (countValidCalls) {
		//	startAudioTunnel(NULL);
		;
	} else {
		//	stopAudioTunnel(NULL);
		;
	}
    if (needRepoll) 
        RIL_requestTimedCallback (sendCallStateChanged, NULL, &TIMEVAL_CALLSTATEPOLL);
    
    return;
error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestDial(void *data, size_t datalen __unused, RIL_Token t)
{
    RIL_Dial *p_dial;
    char *cmd;
    const char *clir;
    int ret;

    p_dial = (RIL_Dial *)data;

    switch (p_dial->clir) {
        case 1: clir = "I"; break;  /*invocation*/
        case 2: clir = "i"; break;  /*suppression*/
        default:
        case 0: clir = ""; break;   /*subscription default*/
    }

    asprintf(&cmd, "ATD%s%s;", p_dial->address, clir);

    ret = at_send_command(cmd, NULL);

    free(cmd);

    /* success or failure is ignored by the upper layer here.
       it will call GET_CURRENT_CALLS and determine success that way */
	if (ret < 0)
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	else
		RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestHangup(void *data, size_t datalen __unused, RIL_Token t)
{
    int ret;
    // 3GPP 22.030 6.5.5
    // "Releases a specific active call X"

	if(CALLS_NUM(sMdmInfo) > 1)
		ret = at_send_command("ATH", NULL);
	else
		ret = at_send_command("AT+CHUP", NULL);

    /* success or failure is ignored by the upper layer here.
       it will call GET_CURRENT_CALLS and determine success that way */
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static int getLteSignalStrength(int *rsrp,int *rsrq)
{
	ATResponse *p_response = NULL;
	int err;
	int skip,rsrp_temp,rsrq_temp;
	char *line;
	int i = 0;
	
	*rsrp = 0x7FFFFFFF;
	*rsrq = 0x7FFFFFFF;
	err = at_send_command_singleline("AT+CESQ", "+CESQ:", &p_response);
	if (err != 0 || p_response->success == 0)
		goto error;
	line = p_response->p_intermediates->line;
	err = at_tok_start(&line);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&skip);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&skip);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&skip);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&skip);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&rsrq_temp);
	if (err < 0) goto error;
	err = at_tok_nextint(&line,&rsrp_temp);
	if (err < 0) goto error;

	//rsrp to real
	if (0 == rsrp_temp){
		*rsrp = 140;
	}else if (1 <= rsrp_temp && 96 >= rsrp_temp){
		*rsrp = 141 - rsrp_temp;
	}else if (97 == rsrp_temp){
		*rsrp = 44;
	}else if (255 == rsrp_temp){
		*rsrp = 0x7FFFFFFF;
	}

	//rsrq to real
    i = rsrq_temp % 2;
    if (255 == rsrq_temp){
        *rsrq = 0x7FFFFFFF;
    }else if (0 <= rsrq_temp && 34 >= rsrq_temp){
        if (0 == i){
            *rsrq = rsrq_temp / 2 - 20;
        }if (1 == i){
            *rsrq = (rsrq_temp + 1) / 2 - 20;
        }
    }
	
	if(*rsrq != 0x7FFFFFFF)
		*rsrq = *rsrq * (-1);

	at_response_free(p_response);
	return 0;
error:
	at_response_free(p_response);
	return -1;
}

static void requestSignalStrength(void *data __unused, size_t datalen __unused, RIL_Token t)
{
	RIL_SignalStrength_v6 signalStrength;
	ATResponse *p_response = NULL;
	int err;
	char *line;
	int ber;
	int rssi,rsrp,rsrq;
	int tech;

	memset(&signalStrength, 0, sizeof(RIL_SignalStrength_v6));

	signalStrength.GW_SignalStrength.signalStrength = 99;
	signalStrength.GW_SignalStrength.bitErrorRate = 99;	
	signalStrength.CDMA_SignalStrength.dbm = -1;
	signalStrength.CDMA_SignalStrength.ecio = -1;
	signalStrength.EVDO_SignalStrength.dbm = -1;	
	signalStrength.EVDO_SignalStrength.ecio = -1;	
	signalStrength.EVDO_SignalStrength.signalNoiseRatio = -1;
	signalStrength.LTE_SignalStrength.signalStrength = -1;
	signalStrength.LTE_SignalStrength.rsrp = 0x7FFFFFFF;
	signalStrength.LTE_SignalStrength.rsrq = 0x7FFFFFFF;
	signalStrength.LTE_SignalStrength.rssnr = 0x7FFFFFFF;
	signalStrength.LTE_SignalStrength.cqi = 0x7FFFFFFF;
	
	err = at_send_command_singleline("AT+CSQ", "+CSQ:", &p_response);
	if (err != 0)
		goto error;

	line = p_response->p_intermediates->line;

	err = at_tok_start(&line);
	if (err < 0) goto error;

	err = at_tok_nextint(&line,&rssi);
	if (err < 0) goto error;

	err = at_tok_nextint(&line, &ber);
	if (err < 0)
		goto error;
	
	RLOGI("SignalStrength %d BER: %d,%d", rssi, ber,tech);
	
	signalStrength.LTE_SignalStrength.signalStrength = 113-2*rssi;
 	
	RIL_onRequestComplete(t, RIL_E_SUCCESS, &signalStrength,sizeof(RIL_SignalStrength_v6));

	at_response_free(p_response);
	return;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);

}

static const char* statusTable(int state)
{
    switch(state){
	case 0: return("unknown");   break;
	case 1: return("available"); break;
	case 2: return("current");   break;
	case 3: return("forbidden"); break;
	default: return NULL;
    }
}

static void requestQueryAvailableNetworks(void *data, size_t datalen, RIL_Token t)
{
    int err, i, skip, operators, status;
    char *p = NULL, * c_skip, *line;
    char ** response = NULL;
    ATResponse *p_response = NULL;
	
    err = at_send_command_singleline("AT+COPS=?", "+COPS:", &p_response);
    if (err != 0) goto error;
	
    line = p_response->p_intermediates->line;
    err = at_tok_start(&line);
    if (err < 0) goto error;
	
    operators = 0;
    for (p = line ; *p != '\0' ;p++) {
		if (*p == '(') operators++;
    }
	
    response = (char **)alloca(operators * 4 * sizeof(char *));
    for (i = 0 ; i < operators ; i++ )
	{
		err = at_tok_nextstr(&line, &c_skip);
		if (err < 0) goto error;
		if (strcmp(c_skip,"") == 0)
		{
		    operators = i;
		    continue;
		}
		status = atoi(&c_skip[1]);
		response[i*4+3] = (char*)statusTable(status);
		err = at_tok_nextstr(&line, &(response[i*4+0]));
		if (err < 0) goto error;
		
		err = at_tok_nextstr(&line, &(response[i*4+1]));
		if (err < 0) goto error;
		
		err = at_tok_nextstr(&line, &(response[i*4+2]));
		if (err < 0) goto error;
		
		err = at_tok_nextstr(&line, &c_skip);
		if (err < 0) goto error;
	}
    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, (operators * 4 * sizeof(char *)));
    at_response_free(p_response);
    return;
error:
    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetPreferredNetworkType( int request __unused, void *data,
                                            size_t datalen __unused, RIL_Token t )
{

	int preferred,newTech,oldTech,err;
	const char* cmd;

	assert (datalen >= sizeof(int *));
	preferred = ((int *)data)[0];

	RLOGI("requestSetPreferredNetworkType: current:%x. New: %x", PREFERRED_NETWORK(sMdmInfo), preferred);	
	
	int old = PREFERRED_NETWORK(sMdmInfo);
	if(old !=preferred )
	{
		switch (preferred) {
		case PREF_NET_TYPE_GSM_WCDMA:		 /* GSM/WCDMA (WCDMA preferred) */
		case PREF_NET_TYPE_GSM_WCDMA_AUTO:  /* GSM/WCDMA (auto mode, according to PRL) */
			cmd = "AT^SYSCONFIG=26,21,1,2";	/* Automatic, prefer WCDMA */
			newTech = MDM_WCDMA;
			break;
		case PREF_NET_TYPE_GSM_ONLY:		 /* GSM only */
			cmd = "AT^SYSCONFIG=13,0,1,2"; /* GSM only */
			newTech = MDM_GSM;
			break;
		case PREF_NET_TYPE_WCDMA:			 /* WCDMA only */
			cmd = "AT^SYSCONFIG=14,0,1,2"; /* WCDMA only */
			newTech = MDM_WCDMA;
			break;
		case PREF_NET_TYPE_CDMA_EVDO_AUTO: /* CDMA and EvDo (auto mode, according to PRL) */
		case PREF_NET_TYPE_CDMA_ONLY:                 /* CDMA only */
		case PREF_NET_TYPE_EVDO_ONLY:                 /* EvDo only */
			RIL_onRequestComplete(t, RIL_E_MODE_NOT_SUPPORTED, NULL, 0);
			newTech = MDM_EVDO;
			return;
		case PREF_NET_TYPE_LTE_CDMA_EVDO:       /* LTE, CDMA and EvDo */
		case PREF_NET_TYPE_LTE_ONLY:                 /* LTE only */
			cmd = "AT^SYSCONFIG=17,0,1,2";
			newTech = MDM_LTE;
			break;
		case PREF_NET_TYPE_LTE_WCDMA:                  /* LTE/WCDMA */
			cmd = "AT^SYSCONFIG=25,0,1,2";
			newTech = MDM_LTE;
			break;
		case PREF_NET_TYPE_LTE_GSM_WCDMA:		 /* LTE, GSM/WCDMA */
		case PREF_NET_TYPE_LTE_CMDA_EVDO_GSM_WCDMA:   /* LTE, CDMA, EvDo, GSM/WCDMA */
		default:
			cmd = "AT+CTEC=2,2";
			newTech = MDM_LTE;
		}

		//if(PPPD_START(sMdmInfo) == PPP_STARTED)
		//onDeactiveDataCallList();

		//err = at_send_command("AT+CFUN=0",NULL);
		//if (err != 0)
			//goto error;
		
		//err = at_send_command(cmd,NULL);
		if (err != 0)
			goto error;


		err = at_send_command("AT+CFUN=1",NULL);
		if (err != 0)
			goto error;

                err = at_send_command("AT+CGDCONT=1,\"IP\",\"CMNET\"",NULL);
		if (err != 0)
			goto error;
		
		//err = at_send_command("AT+CGACT=1,1",NULL);
		//err = at_send_command("AT+CSTT=\"CMNET\",\"\",\"\"",NULL);
		RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, sizeof(int));

		PREFERRED_NETWORK(sMdmInfo) = preferred;
		oldTech = TECH(sMdmInfo);
		RLOGD("requestSetPreferredNetworkType:Tech change (%d => %d)", oldTech, newTech);
		if (newTech != oldTech) {
			TECH(sMdmInfo) = newTech;
			if (techFromModemType(newTech) != techFromModemType(oldTech)) 
			{
				int tech = techFromModemType(TECH(sMdmInfo));
				if (tech > 0 ) {
					RIL_onUnsolicitedResponse(RIL_UNSOL_VOICE_RADIO_TECH_CHANGED,&tech, sizeof(tech));
					return;
					}
				}
		}
	}else{
		goto error;
	}

	return;
error:
	RLOGE("ERROR: requestSetPreferredNetworkType() failed\n");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	 
}

static void requestGetPreferredNetworkType(int request __unused, void *data __unused, size_t datalen __unused, RIL_Token t)
{

	ATResponse *p_response = NULL;
	int err;
	char *line;
	int skip,sys_mode;
	int response = PREF_NET_TYPE_LTE_GSM_WCDMA;

	err = at_send_command_singleline("AT^SYSINFO", "^SYSINFO:", &p_response);

	if (err != 0) {
		// assume radio is off
		goto error;
	}

	line = p_response->p_intermediates->line;

	err = at_tok_start(&line);
	if (err < 0) goto error;

	err = at_tok_nextint(&line, &skip);
	if (err < 0) goto error;

	err = at_tok_nextint(&line, &skip);
	if (err < 0) goto error;
	
	err = at_tok_nextint(&line, &skip);
	if (err < 0) goto error;

	err = at_tok_nextint(&line, &sys_mode);
	if (err < 0) goto error;

	/* Based on reported syscfg */
	if (sys_mode == 3) {
		response = PREF_NET_TYPE_GSM_ONLY;	/* GSM only */
	} else if (sys_mode == 5 || sys_mode == 15) {
		response = PREF_NET_TYPE_WCDMA; 	/* WCDMA only */
	} else if (sys_mode == 17) {
		response = PREF_NET_TYPE_LTE_GSM_WCDMA; /* GSM or WCDMA */
	} else {
		response = PREF_NET_TYPE_LTE_GSM_WCDMA; /* Auto */
	}

	RLOGD("requestGetPreferredNetworkType() mode:%d\n",response);
	RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(int));
	at_response_free(p_response);
	return;

error:
	RLOGE("ERROR: requestGetPreferredNetworkType() failed - modem does not support command\n");
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}


static int parseRegistrationState(char *str, int *items, int **response)
{
    int err;
    char *line = str, *p;
    int *resp = NULL;
    int skip;
    int count = 3;
    int commas;

    RLOGD("parseRegistrationState. Parsing: %s",str);
    err = at_tok_start(&line);
    if (err < 0) goto error;

    /* Ok you have to be careful here
     * The solicited version of the CREG response is
     * +CREG: n, stat, [lac, cid]
     * and the unsolicited version is
     * +CREG: stat, [lac, cid]
     * The <n> parameter is basically "is unsolicited creg on?"
     * which it should always be
     *
     * Now we should normally get the solicited version here,
     * but the unsolicited version could have snuck in
     * so we have to handle both
     *
     * Also since the LAC and CID are only reported when registered,
     * we can have 1, 2, 3, or 4 arguments here
     *
     * finally, a +CEREG: answer may have a fifth value that corresponds
     * to the network type, as in;
     *
     *   +CEREG: n, stat [,lac, cid [,networkType]]
     */

    /* count number of commas */
    commas = 0;
    for (p = line ; *p != '\0' ;p++) {
        if (*p == ',') commas++;
    }

    if(commas > 3)
	count = commas;

    resp = (int *)calloc(count, sizeof(int));
    if (!resp) goto error;
    switch (commas) {
        case 0: /* +CREG: <stat> */
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            resp[1] = -1;
            resp[2] = -1;
        	break;

        case 1: /* +CREG: <n>, <stat> */
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            resp[1] = -1;
            resp[2] = -1;
            if (err < 0) goto error;
        	break;

        case 2: /* +CREG: <stat>, <lac>, <cid> */
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[2]);
            if (err < 0) goto error;
        	break;
        case 3: /* +CREG: <n>,<stat>, <lac>, <cid>*/
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[2]);
            if (err < 0) goto error;
        	break;
        /* special case for CEREG, there is a fourth parameter
         * that is the network type (unknown/gprs/edge/umts)
         */
         		
        case 4: /* +CREG: <n>, <stat>, <lac>, <cid>, <act> */   /////////* +CEREG: <stat>, <lac>, <cid>, <act> <SubAct>*/
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[2]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[3]);
            if (err < 0) goto error;
        	break;
		case 5:	 /* +CEREG: <n>, <stat>, <lac>, <cid>, <act>,<SubAct>  */	
            err = at_tok_nextint(&line, &skip);
            if (err < 0) goto error;
            err = at_tok_nextint(&line, &resp[0]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[1]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[2]);
            if (err < 0) goto error;
            err = at_tok_nexthexint(&line, &resp[3]);
            if (err < 0) goto error;
	   	 	err = at_tok_nexthexint(&line, &resp[4]);
            if (err < 0) goto error;
			break;
        default:
            goto error;
    }
	
    s_lac = resp[1];
    s_cid = resp[2];
		
    if (response)
        *response = resp;
    if (items)
        *items = count;
   // if (type)
       // *type = techFromModemType(TECH(sMdmInfo));
    return 0;
error:
    free(resp);
    return -1;
}

static int convertActToRadioTech(int AcT)
{
	int ret = -1;
	switch (AcT) {
        case 0:
        case 1:
        	ret = RADIO_TECH_GPRS;
        	break;

        case 2:
        	ret = RADIO_TECH_UMTS;
       	 	break;

        case 3:
        	ret = RADIO_TECH_EDGE;
        	break;

        case 4:
       	 	ret = RADIO_TECH_HSDPA;
        	break;

        case 5:
        	ret = RADIO_TECH_HSUPA;
        	break;

        case 6:
        	ret = RADIO_TECH_HSPA;
        break;
        /*LTE*/
        case 7:
        	ret = RADIO_TECH_LTE;
        	break;

        case -1:
        default:
        	ret = RADIO_TECH_UNKNOWN;
        	break;
    }

    return ret;
}


#define REG_STATE_LEN 15
#define REG_DATA_STATE_LEN 7
static void requestRegistrationState(int request, void *data __unused, size_t datalen __unused, RIL_Token t)
{
    int err;
    int *registration;
    char **responseStr = NULL;
    ATResponse *p_response = NULL;
    const char *cmd;
    const char *prefix;
    char *line;
    int i = 0, j, numElements = 0;
    int count = 3;
    int startfrom,radiotech;

    RLOGD("requestRegistrationState");
    if (request == RIL_REQUEST_VOICE_REGISTRATION_STATE) {
        cmd = "AT+CREG?";
        prefix = "+CREG:";
        numElements = REG_STATE_LEN;
    } else if (request == RIL_REQUEST_DATA_REGISTRATION_STATE) {
	if(TECH(sMdmInfo)  == MDM_LTE){
		cmd = "AT+CEREG?";
        	prefix = "+CEREG:";
	}
	else{
		cmd = "AT+CGREG?";
        	prefix = "+CGREG:";
	}
        numElements = REG_DATA_STATE_LEN;
    } else {
        assert(0);
        goto error;
    }

    err = at_send_command_singleline(cmd, prefix, &p_response);

    if (err != 0) goto error;

    line = p_response->p_intermediates->line;

    if (parseRegistrationState(line, &count, &registration)) goto error;

   //numElements = count;
   
    responseStr = malloc(numElements * sizeof(char *));
    if (!responseStr) goto error;
    memset(responseStr, 0, numElements * sizeof(char *));   //clear 0
    /**
     * The first '4' bytes for both registration states remain the same.
     * But if the request is 'DATA_REGISTRATION_STATE',
     * the 5th and 6th byte(s) are optional.
     */
	//for(i = 0; i < count;i++)
	//{	
    	//RLOGD("registration[%d]:%d",i,registration[i]);
	//}

 	// type == RADIO_TECH_3GPP
	//RLOGD("registration state type: 3GPP");
	asprintf(&responseStr[0], "%d", registration[0]);
	asprintf(&responseStr[1], "%x", registration[1]);
	asprintf(&responseStr[2], "%x", registration[2]);
	if (count > 3){
		radiotech = convertActToRadioTech(registration[3]);
		asprintf(&responseStr[3], "%d", radiotech);
	}
	if(count ==5)
		asprintf(&responseStr[4], "%x", registration[4]);

    /**
     * Optional bytes for DATA_REGISTRATION_STATE request
     * 4th byte : Registration denial code
     * 5th byte : The max. number of simultaneous Data Calls
     */
    //if(request == RIL_REQUEST_DATA_REGISTRATION_STATE) {
        // asprintf(&responseStr[4], "3");
        // asprintf(&responseStr[5], "1");
   //}

    //for (j = 0; j < numElements; j++) {
       // if (!resporesponseStrnseStr[i]) goto error;
		//RLOGD("responseStr[%d]=%s",j,responseStr[j]);
    //}
	 
    free(registration);
    registration = NULL;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, responseStr, numElements*sizeof(responseStr));
    for (j = 0; j < numElements; j++ ) {
        free(responseStr[j]);
        responseStr[j] = NULL;
    }
    free(responseStr);
    responseStr = NULL;
    at_response_free(p_response);
   RLOGD("requestRegistrationState ending");
    return;
error:
    if (responseStr) {
        for (j = 0; j < numElements; j++) {
            free(responseStr[j]);
            responseStr[j] = NULL;
        }
        free(responseStr);
        responseStr = NULL;
    }
    RLOGE("requestRegistrationState must never return an error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestOperator(void *data __unused, size_t datalen __unused, RIL_Token t)
{
    int err;
    int i;
    int skip;
    ATLine *p_cur;
    char *response[3];

    memset(response, 0, sizeof(response));

    ATResponse *p_response = NULL;

    err = at_send_command_multiline(
        "AT+COPS=3,0;+COPS?;+COPS=3,1;+COPS?;+COPS=3,2;+COPS?",
        "+COPS:", &p_response);

    /* we expect 3 lines here:
     * +COPS: 0,0,"T - Mobile"
     * +COPS: 0,1,"TMO"
     * +COPS: 0,2,"310170"
     */

    if (err != 0) goto error;

    for (i = 0, p_cur = p_response->p_intermediates
            ; p_cur != NULL
            ; p_cur = p_cur->p_next, i++
    ) {
        char *line = p_cur->line;

        err = at_tok_start(&line);
        if (err < 0) goto error;

        err = at_tok_nextint(&line, &skip);
        if (err < 0) goto error;

        // If we're unregistered, we may just get
        // a "+COPS: 0" response
        if (!at_tok_hasmore(&line)) {
            response[i] = NULL;
            continue;
        }

        err = at_tok_nextint(&line, &skip);
        if (err < 0) goto error;

        // a "+COPS: 0, n" response is also possible
        if (!at_tok_hasmore(&line)) {
            response[i] = NULL;
            continue;
        }

        err = at_tok_nextstr(&line, &(response[i]));
        if (err < 0) goto error;
        // Simple assumption that mcc and mnc are 3 digits each
        if (strlen(response[i]) == 5) {
            if (sscanf(response[i], "%3d%2d", &s_mcc, &s_mnc) != 2) {
                RLOGE("requestOperator expected mccmnc to be 6 decimal digits");
            }
        }
    }

    if (i != 3) {
        /* expect 3 lines exactly */
        goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    at_response_free(p_response);

    return;
error:
    RLOGE("requestOperator must not return error when radio is on");
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}


wait_pdu* dequeuePdu(void)
{
    wait_pdu* wpdu = NULL;
    if (NULL != g_wait_pdu)
    {
        wpdu = g_wait_pdu;
        g_wait_pdu = wpdu->next;
        wpdu->next = NULL;
    }
    return wpdu;
}

void enqueuePdu(int type, const char* sms_pdu)
{
   
    wait_pdu* wpdu = malloc(sizeof(wait_pdu));
    if (NULL == wpdu)
    {
        RLOGD("%s() malloc failed", __FUNCTION__);
        return;
    }
    memset(wpdu, 0, sizeof(wait_pdu));
    wpdu->type = type;
    wpdu->pdu  = strdup(sms_pdu);
    if (NULL == wpdu->pdu)
    {
        RLOGD("%s() strdup sms pdu failed", __FUNCTION__);
        free(wpdu);  //coverity=RESOURCE_LEAK
        return;
    }

    if (NULL == g_wait_pdu)
    {
        g_wait_pdu = wpdu;
    }
    else
    {
        wait_pdu* p = g_wait_pdu;
        while (NULL != p->next)
        {
            p = p->next;
        }
        p->next = wpdu;
    }
}


static void requestSendSMS(void *data, size_t datalen, RIL_Token t)
{
    int err;
    const char *smsc;
    const char *pdu;
    int tpLayerLength;
    char *cmd1, *cmd2;
    RIL_SMS_Response response;
    ATResponse *p_response = NULL;
	char *line;

    memset(&response, 0, sizeof(response));
    RLOGD("requestSendSMS datalen =%zu", datalen);

    smsc = ((const char **)data)[0];
    pdu = ((const char **)data)[1];

    tpLayerLength = strlen(pdu)/2;

    // "NULL for default SMSC"
    if (smsc == NULL) {
        smsc= "00";
    }

    asprintf(&cmd1, "AT+CMGS=%d", tpLayerLength);
    asprintf(&cmd2, "%s%s", smsc, pdu);

    err = at_send_command_sms(cmd1, cmd2, "+CMGS:", &p_response);
	free(cmd1);
	free(cmd2);
	
    if (err < 0 ) goto error;
	if(p_response->success == 0) goto cms_error;

	line = p_response->p_intermediates->line;
    memset(&response, 0, sizeof(response));

    err = at_tok_start(&line);
    err = at_tok_nextint(&line, &(response.messageRef));
    if (at_tok_hasmore(&line))
    {
        err = at_tok_nextstr(&line, &(response.ackPDU));
    }
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(response));
    at_response_free(p_response);

    return;
error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
    return;
cms_error:
    // send retry error.
    RIL_onRequestComplete(t, RIL_E_SMS_SEND_FAIL_RETRY, NULL, 0);
    at_response_free(p_response);
    return;
    }

static void requestSendSMSExpectMore(void *data, size_t datalen, RIL_Token t)
{
    /* Throw the command on the channel and ignore any errors, since we
       need to send the SMS anyway and subsequent SMSes will be sent anyway. */
    at_send_command("AT+CMMS=1",NULL);

    requestSendSMS(data, datalen, t);
}

static void requestWriteSmsToSim(void *data, size_t datalen __unused, RIL_Token t)
{
	RIL_SMS_WriteArgs *args;
	char *cmd,*pdu;
	char *line;
	int length,index;
	int err;
	ATResponse *p_response = NULL;

	(void) datalen;

	args = (RIL_SMS_WriteArgs *) data;

	length = strlen(args->pdu) / 2;

	asprintf(&cmd, "AT+CMGW=%d,%d", length, args->status);
	asprintf(&pdu, "%s%s", (args->smsc ? args->smsc : "00"), args->pdu);
	err = at_send_command_sms(cmd, pdu, "+CMGW:", &p_response);
	free(cmd);
	free(pdu);

	if (err != 0)
		goto error;

	line = p_response->p_intermediates->line;

	err = at_tok_start(&line);
	if (err < 0)
		goto error;

	err = at_tok_nextint(&line, &index);
	if (err < 0)
		goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, &index, sizeof(int *));
	at_response_free(p_response);
	return;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}


static void requestDeleteSmsOnSim(void *data, size_t datalen, RIL_Token t)
{
    int err;
	char *cmd;
	/* cmiot modems use a 0-based message slot index, but the SIM record is 1-based. 
	   We will translate Huawei indices to Record indices, to make the Android RIL 
	   able to read SMS stored in SIM using SimIo with the same indices as the Sms 
	   delete command and NewSmsOnSim index */
	int idx = ((int *) data)[0] - 1;
	if (idx < 0) {
		RLOGE("DeleteSmsOnSim: Invalid index! (%d)",idx);
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
		return;
	}

	asprintf(&cmd,"AT+CMGD=%d", idx);
    err = at_send_command(cmd,NULL);
	free(cmd);
    if (err != 0)
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    else
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}


static void requestGetSMSCAddress(RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err;
    char *line;
    char *response;

    err = at_send_command_singleline("AT+CSCA?", "+CSCA:", &p_response);

    if (err != 0)
        goto error;

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0)
        goto error;

    err = at_tok_nextstr(&line, &response);
    if (err < 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(char *));
    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}

/**
 * RIL_REQUEST_SET_SMSC_ADDRESS
 */
static void requestSetSMSCAddress(void *data, size_t datalen, RIL_Token t)
{
    (void) datalen;
    int err;	
	char *cmd;
    const char *smsc = (const char *)data;

	asprintf(&cmd,"AT+CSCA=\"%s\"", smsc);
    err = at_send_command(cmd,NULL);
	free(cmd);
    if (err != 0)
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    else
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

/**
 * RIL_REQUEST_REPORT_SMS_MEMORY_STATUS
 */
static void requestSmsStorageStatus(void *data, size_t datalen, RIL_Token t)
{
    int ack;

    ack = ((int *) data)[0];

    /* Android will call RIL_REQUEST_REPORT_SMS_MEMORY_STATUS in case of:
     * 0. memory is full
     * 1. memory was full and been cleaned up, inform modem memory is available now.
     */
    switch (ack) {
    case 0:
        /* Android will handle this, no need to inform modem. always return success. */
        RLOGI("SMS storage full");
        break;

    case 1:
        /* Since we are not using +CNMA command. It's fine to return without informing network */
        RLOGI("Failed to inform network for Message Cleanup. Need cmd : ESMSMEMAVAIL");
        break;

    default:
        RLOGE("%s() Invalid parameter", __func__);
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestSMSAcknowledge(void *data, size_t datalen __unused, RIL_Token t)
{
    int ackSuccess;
    int err;
	wait_pdu* wpdu = NULL;
	
    ackSuccess = ((int *)data)[0];

    if (ackSuccess == 1) {
        err = at_send_command("AT+CNMA=1", NULL);
    } else if (ackSuccess == 0)  {
        err = at_send_command("AT+CNMA=2", NULL);
    } else {
        RLOGE("unsupported arg to RIL_REQUEST_SMS_ACKNOWLEDGE\n");
        goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);

	pthread_mutex_lock(&s_pdu_mutex);
    wpdu = dequeuePdu();
    if (NULL != wpdu)
    {
        int unsol_response = 0;
        if (OUTSTANDING_CMT_SMS == wpdu->type)
        {
            unsol_response = RIL_UNSOL_RESPONSE_NEW_SMS;
        }
        else if (OUTSTANDING_CDS_REPORT == wpdu->type)
        {
            unsol_response = RIL_UNSOL_RESPONSE_NEW_SMS_STATUS_REPORT;
        }

        RIL_onUnsolicitedResponse(unsol_response, wpdu->pdu, strlen(wpdu->pdu));
        free(wpdu->pdu);
        free(wpdu);
    }
    else
    {
        g_outstanding_sms = 0;
    }
    pthread_mutex_unlock(&s_pdu_mutex);

	return;
	
error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

#define BROADCAST_MAX_RANGES_SUPPORTED 10

/**
 * RIL_REQUEST_GSM_GET_BROADCAST_SMS_CONFIG
 */
static void requestGSMGetBroadcastSMSConfig(RIL_Token t)
{
    ATResponse *p_response = NULL;
    int mode, err = 0;
    unsigned int i, count = 0;
    char *mids;
    char *range;
    char *trange;
    char *tok = NULL;

    RIL_GSM_BroadcastSmsConfigInfo *configInfo[BROADCAST_MAX_RANGES_SUPPORTED];

    err = at_send_command_singleline("AT+CSCB?", "+CSCB:", &p_response);

    if (err != 0)
        goto error;

    tok = p_response->p_intermediates->line;

    err = at_tok_start(&tok);
    if (err < 0)
        goto error;

    err = at_tok_nextint(&tok, &mode);
    if (err < 0)
        goto error;

    /**
     * Get the string that yields the service ids (mids). AT+CSCB <mids>
     * parameter may contain a mix of single service ids (,%d,) and service id
     * ranges (,%d-%d,).
     */
    err = at_tok_nextstr(&tok, &mids);
    if (err < 0)
        goto error;

    while (at_tok_nextstr(&mids, &range) == 0) {
        /**
         * Replace any '-' sign with ',' sign to allow for at_tok_nextint
         * for both fromServiceId and toServiceId below.
         */
        trange = range;
        while ((NULL != trange) && ('\0' != *trange)) {
            if ('-' == *trange)
                *trange = ',';
            trange++;
        }
        if (count < BROADCAST_MAX_RANGES_SUPPORTED) {
            configInfo[count] = (RIL_GSM_BroadcastSmsConfigInfo*) calloc(1,
                sizeof(RIL_GSM_BroadcastSmsConfigInfo));
            if (NULL == configInfo[count])
                goto error;

            /* No support for "Not accepted mids", selected is always 1 */
            configInfo[count]->selected = 1;

            /* Fetch fromServiceId value */
            err = at_tok_nextint(&range, &(configInfo[count]->fromServiceId));
            if (err < 0)
                goto error;
            /* Try to fetch toServiceId value if it exist */
            err = at_tok_nextint(&range, &(configInfo[count]->toServiceId));
            if (err < 0)
                configInfo[count]->toServiceId =
                    configInfo[count]->fromServiceId;

            count++;
        } else {
            RLOGW("%s() Max limit (%d) passed, can not send all ranges "
                 "reported by modem.", __func__,
                 BROADCAST_MAX_RANGES_SUPPORTED);
            break;
        }
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &configInfo,sizeof(RIL_GSM_BroadcastSmsConfigInfo *) * count);

    goto exit;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);

exit:
    at_response_free(p_response);
    for (i = 0; i < count; i++)
        free(configInfo[i]);
}



/**
 * RIL_REQUEST_GSM_SET_BROADCAST_SMS_CONFIG
 */
static void requestGSMSetBroadcastSMSConfig(void *data, size_t datalen, RIL_Token t)
{
    int err, count, i;
    char *tmp, *mids = NULL;
	char *cmd;
    RIL_GSM_BroadcastSmsConfigInfo **configInfoArray =
        (RIL_GSM_BroadcastSmsConfigInfo **) data;
    RIL_GSM_BroadcastSmsConfigInfo *configInfo = NULL;

    count = datalen / sizeof(RIL_GSM_BroadcastSmsConfigInfo *);
    RLOGI("Number of MID ranges in BROADCAST_SMS_CONFIG: %d", count);

    for (i = 0; i < count; i++) {
        configInfo = configInfoArray[i];
        /* No support for "Not accepted mids" in AT */
        if (configInfo->selected) {
            tmp = mids;
            asprintf(&mids, "%s%d-%d%s", (tmp ? tmp : ""),
                configInfo->fromServiceId, configInfo->toServiceId,
                (i == (count - 1) ? "" : ",")); /* Last one? Skip comma */
            free(tmp);
        }
    }

    if (mids == NULL) {
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

	asprintf(&cmd,"AT+CSCB=0,\"%s\"", mids);
    err = at_send_command(cmd,NULL);
    free(mids);
	free(cmd);

    if (err != 0) {
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

/**
 * RIL_REQUEST_GSM_SMS_BROADCAST_ACTIVATION
 */
static void requestGSMSMSBroadcastActivation(void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response = NULL;
    int mode, mt, bm, ds, bfr, skip;
    int activation;
    char *tok;
    int err;
	char *cmd;

    (void) datalen;

    /* AT+CNMI=[<mode>[,<mt>[,<bm>[,<ds>[,<bfr>]]]]] */
    err = at_send_command_singleline("AT+CNMI?", "+CNMI:", &p_response);
    if (err != 0)
        goto error;

    tok = p_response->p_intermediates->line;

    err = at_tok_start(&tok);
    if (err < 0)
        goto error;
    err = at_tok_nextint(&tok, &mode);
    if (err < 0)
        goto error;
    err = at_tok_nextint(&tok, &mt);
    if (err < 0)
        goto error;
    err = at_tok_nextint(&tok, &skip);
    if (err < 0)
        goto error;
    err = at_tok_nextint(&tok, &ds);
    if (err < 0)
        goto error;
    err = at_tok_nextint(&tok, &bfr);
    if (err < 0)
        goto error;

    /* 0 - Activate, 1 - Turn off */
    activation = *((const int *)data);
    if (activation == 0)
        bm = 2;
    else
        bm = 0;

	asprintf(&cmd,"AT+CNMI=%d,%d,%d,%d,%d", mode, mt, bm, ds, bfr);
    err = at_send_command(cmd,NULL);
	free(cmd);
    if (err != 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);

    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void newSmsOrStatusReportOnSIM(const char *s)
{
	int err = 0;
	int index = -1;
	char* line = NULL;
	char* linefree = NULL;
	char* psz_sms_mem = NULL;
	//char cmd[2];

	line = strdup(s);
	linefree = line;

	err = at_tok_start(&line);
	if (err < 0){
		goto error;
	}

	err = at_tok_nextstr(&line, &psz_sms_mem);
	if (err < 0){
		goto error;
	}

	if (0 != strncmp(psz_sms_mem, "SM", 2)){
		goto error;
	}

	err = at_tok_nextint(&line, &index);
	if (err < 0){
		goto error;
	}

	RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_NEW_SMS_ON_SIM,&index, sizeof(int *));

error:
	if (NULL != linefree)
		free(linefree);
}

static void newSmsReceived(const char *sms_pdu)
{
    pthread_mutex_lock(&s_pdu_mutex);
    if (g_outstanding_sms)
    {
        /* No RIL_UNSOL_RESPONSE_NEW_SMS or RIL_UNSOL_RESPONSE_NEW_SMS_STATUS_REPORT
         * messages should be sent until a RIL_REQUEST_SMS_ACKNOWLEDGE has been received for
         * previous new SMS.
         */
        RLOGD("Waiting for acknowledge for previous sms, enqueueing PDU");
        enqueuePdu(OUTSTANDING_CMT_SMS, sms_pdu);
    }
    else
    {
        g_outstanding_sms = 1;
        RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_NEW_SMS, sms_pdu, strlen(sms_pdu));
    }
    pthread_mutex_unlock(&s_pdu_mutex);
}

static void newSmsStatusReportReceived(const char *sms_pdu)
{
	pthread_mutex_lock(&s_pdu_mutex);
	if (g_outstanding_sms)
	{
		/* No RIL_UNSOL_RESPONSE_NEW_SMS or RIL_UNSOL_RESPONSE_NEW_SMS_STATUS_REPORT
		 * messages should be sent until a RIL_REQUEST_SMS_ACKNOWLEDGE has been received for
		 * previous new SMS.
		 */
		RLOGD("Waiting for acknowledge for previous sms, enqueueing PDU");
		enqueuePdu(OUTSTANDING_CDS_REPORT, sms_pdu);
	}
	else
	{
		g_outstanding_sms = 1;
		RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_NEW_SMS_STATUS_REPORT, sms_pdu, strlen(sms_pdu));
	}
	pthread_mutex_unlock(&s_pdu_mutex);
}

static void newBroadcastSms(const char *sms_pdu)
{
 	RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_NEW_BROADCAST_SMS, sms_pdu, strlen(sms_pdu));
}

static void simSmsStorageFull(const char *s,const char *sms_pdu)
{
	int err;
	char	*line = NULL;
	char	*linesave = NULL;

	int reason = 0;
	line = strdup(s);
	linesave = line;
	err = at_tok_start(&line);

	err = at_tok_nextint(&line, &reason);

	if (reason == 211 || reason == 322)
	{
		RIL_onUnsolicitedResponse(RIL_UNSOL_SIM_SMS_STORAGE_FULL, NULL, 0);
	}
error:
	if (linesave != NULL){
		free(linesave);
	}
	RLOGW("%s: Error parameter in ind msg: %s", __FUNCTION__, s);
	return;
}

#if 0
static int writelineEx (int fd, const char *s)
{
    size_t cur = 0;
    size_t len = strlen(s);
    ssize_t written;

    if (fd < 0 ) {
        return AT_ERROR_CHANNEL_CLOSED;
    }

    RLOGD("AT> %s\n", s);

    AT_DUMP( ">> ", s, strlen(s) );

    /* the main string */
    while (cur < len) {
        do {
            written = write (fd, s + cur, len - cur);
        } while (written < 0 && errno == EINTR);

        if (written < 0) {
            return AT_ERROR_GENERIC;
        }

        cur += written;
    }

    /* the \r  */

    do {
        written = write (fd, "\r" , 1);
    } while ((written < 0 && errno == EINTR) || (written == 0));

    if (written < 0) {
        return AT_ERROR_GENERIC;
    }

    return 0;
}
#endif

static int getAddressInfo(void)
{
	char local_ip[PROPERTY_VALUE_MAX] = {'\0'};
	char dnses[2*ADDRESS_LEN+1] = {'\0'};
	char dns1[PROPERTY_VALUE_MAX] = {'\0'};
	char dns2[PROPERTY_VALUE_MAX] = {'\0'};
	char gw[PROPERTY_VALUE_MAX] = {'\0'};	

	if(USE_PPP(sMdmInfo)){
		// get local_ip
		memset(rmnetParam.ip_addr,0,sizeof(rmnetParam.ip_addr));
		property_get("net.ppp0.local-ip", local_ip, NULL);
		if(strcmp(local_ip, "") != 0){ 
			strncpy(rmnetParam.ip_addr,local_ip,sizeof(rmnetParam.ip_addr) - 1);
		}

		//get dns
		memset(rmnetParam.dns_addr,0,sizeof(rmnetParam.dns_addr));
		property_get("net.ppp0.dns1", dns1, NULL);
		property_get("net.ppp0.dns2", dns2, NULL);
		snprintf(dnses,sizeof(dnses) - 1, "%s %s", dns1, dns2);
		if((strcmp(dns1, "") != 0) && (strcmp(dns2, "") != 0)){		
			strncpy(rmnetParam.dns_addr,dnses,sizeof(rmnetParam.dns_addr) - 1);
		}
		else{
			strcpy(rmnetParam.dns_addr,"8.8.8.8 8.8.4.4");
		}

		// get gateway
		memset(rmnetParam.gateways,0,sizeof(rmnetParam.gateways));
		property_get("net.ppp0.remote-ip", gw, NULL);
		if(strcmp(gw, "") != 0){ 
			strncpy(rmnetParam.gateways,gw,sizeof(rmnetParam.gateways) - 1);
		}

		RLOGD("local_ip=%s,dnses=%s,gw=%s",local_ip,dnses,gw);
		
		if(strstr(local_ip,"192.168") != NULL)
			return -1;
		
		return 0;
	}

	return -1;
}

#define PROCNET_ROUTE_PATH			"/proc/net/route"
#define PROCNET_ROUTE_IPV6_PATH		"/proc/net/ipv6_route"
#define PROCNET_IFINET6_PATH		"/proc/net/if_inet6"
#define SCOPE_LINK_STR				"fe80"
#define TYPE_IPV4					0
#define TYPE_IPV6					1

//failed -1;success 0
static int get_ipv4_route(const char *ifname, char *addr)
{
    char name[64];
	char *tmpaddr;
    in_addr_t dest, gway, mask;
    int flags, refcnt, use, metric, mtu, win, irtt;
	struct in_addr mygateway;
    FILE *fp;

	if (!ifname || !addr)
		return -1;

    fp = fopen(PROCNET_ROUTE_PATH, "r");
    if (fp == NULL)
        return -1;
    /* Skip the header line */
    if (fscanf(fp, "%*[^\n]\n") < 0) {
        fclose(fp);
        return -1;
    }
    ifc_init();
	mygateway.s_addr = 0;
    for (;;) {
        int nread = fscanf(fp, "%63s%X%X%X%d%d%d%X%d%d%d\n",
                           name, &dest, &gway, &flags, &refcnt, &use, &metric, &mask,
                           &mtu, &win, &irtt);
        if (nread != 11) {
            break;
        }
        if ((flags & (RTF_UP|RTF_GATEWAY)) == (RTF_UP|RTF_GATEWAY)
                && dest == 0
                && strcmp(ifname, name) == 0) {
            mygateway.s_addr = gway;
			tmpaddr = inet_ntoa(mygateway);
			strncpy(addr, tmpaddr, strlen(tmpaddr));
			fclose(fp);
    		ifc_close();
			//RLOGD("get_ipv4_route:addr=%s", addr);
            return 0;
        }
    }
    fclose(fp);
    ifc_close();
    return -1;
}

//failed -1;success 0
int get_ipv6_local_addr(char type, const char *ifname, char *addr, int length)
{
	int ret = 0;
	FILE *fp;
	char addr6[46] = {0}, dev_name[20] = {0};
	char addr6p[8][5];
	int length_of_prefix, scope_value, if_flags, if_index;
	struct sockaddr_in6 sap = {0};
	
	if (!ifname || !addr || length <= 0)
		return -1;
 
	fp = fopen(PROCNET_IFINET6_PATH, "r");
	if (!fp) return -1;

 	ifc_init6();
	//lockf(fileno(fp), F_LOCK, 0);	
	while (fscanf(fp, "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
		addr6p[0], addr6p[1], addr6p[2], addr6p[3],
		addr6p[4], addr6p[5], addr6p[6], addr6p[7],
		 &if_index, &length_of_prefix, &scope_value, &if_flags, dev_name) != EOF) {
		if (!strcmp(dev_name, ifname) && if_flags == 00) {
			sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
				addr6p[0], addr6p[1], addr6p[2], addr6p[3],
				addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
			if (strncmp(addr6p[0], SCOPE_LINK_STR, strlen(SCOPE_LINK_STR)) == 0) {
				//fe80 Scope:Link
				if (type == 1) {
					inet_pton(AF_INET6, addr6, &sap.sin6_addr);
					inet_ntop(AF_INET6, &sap.sin6_addr, addr, length);
					//snprintf(prefix, length2, "%d", length_of_prefix);
					//lockf(fileno(fp), F_ULOCK, 0);
					fclose(fp);
					ifc_close6();
					return 0;
				}
			} else {
				//2001 Scope:Global
				if (type == 0) {
					inet_pton(AF_INET6, addr6, &sap.sin6_addr);
					inet_ntop(AF_INET6, &sap.sin6_addr, addr, length);
					//snprintf(prefix, length2, "%d", length_of_prefix);
					//lockf(fileno(fp), F_ULOCK, 0);
					fclose(fp);
					ifc_close6();
					return 0;
				}
			}
		}
	}
	//lockf(fileno(fp), F_ULOCK, 0);	
	fclose(fp);
	ifc_close6();
	return -1;
}

//failed -1;success 0
static int get_ipv6_route(char type, const char *ifname, char *addr, int length)
{
    FILE *fp;
    char addr6[46] = {0};
    char tmpAdd6[8][5];
    char tmpOther[9][64];
    struct sockaddr_in6 sap = {0};
    
    if (!ifname || !addr || length <= 0)
        return -1;
 
    fp = fopen(PROCNET_ROUTE_IPV6_PATH, "r");
    if (!fp) return -1;
    
	ifc_init6();
    //lockf(fileno(fp), F_LOCK, 0);
    while (fscanf(fp, "%s %02s %s %s %4s%4s%4s%4s%4s%4s%4s%4s %s %s %s %s %20s\n",
        tmpOther[0], tmpOther[1], tmpOther[2], tmpOther[3],
        tmpAdd6[0], tmpAdd6[1], tmpAdd6[2], tmpAdd6[3],
        tmpAdd6[4], tmpAdd6[5], tmpAdd6[6], tmpAdd6[7],
        tmpOther[4], tmpOther[5], tmpOther[6], tmpOther[7], tmpOther[8]) != EOF) {
    	
        if (!strcmp(tmpOther[8], ifname) && atoi(tmpOther[1]) == 0) {
            sprintf(addr6, "%s:%s:%s:%s:%s:%s:%s:%s",
                tmpAdd6[0], tmpAdd6[1], tmpAdd6[2], tmpAdd6[3],tmpAdd6[4], tmpAdd6[5], tmpAdd6[6], tmpAdd6[7]);			
				if (strncmp(tmpAdd6[0], SCOPE_LINK_STR, strlen(SCOPE_LINK_STR)) == 0) {
				//fe80 Scope:Link
				if (type == 1) {
                    inet_pton(AF_INET6, addr6, &sap.sin6_addr);
                    inet_ntop(AF_INET6, &sap.sin6_addr, addr, length);
                    //lockf(fileno(fp), F_ULOCK, 0);
                    fclose(fp);
					ifc_close6();
                    return 0;
                }
            } else {
                //2001 Scope:Global
                if (type == 0) {
                    inet_pton(AF_INET6, addr6, &sap.sin6_addr);
                    inet_ntop(AF_INET6, &sap.sin6_addr, addr, length);
                    //lockf(fileno(fp), F_ULOCK, 0);
                    fclose(fp);
					ifc_close6();
                    return 0;
                }
            }
        }
    }
    //lockf(fileno(fp), F_ULOCK, 0);
    fclose(fp);
	ifc_close6();
    return -1;
}

//failed -1;success 0
static int RIL_ifc_get_default_route(const char *ifname, char *addr, int length, int type)
{
	int err;
	if(type == TYPE_IPV4)
		err = get_ipv4_route(ifname, addr);
	else	
		err = get_ipv6_route(1, ifname, addr, length);

	//RLOGD("err=%d,RIL_ifc_get_default_route:addr=%s", err,addr);
	return err;
}

//failed -1;success 0
static int formatDNSAddr(char **p_cur, char *p_out1, char *p_out2, int length1)
{
	int i = 0;
	int j = 0;
	int ip1[32];
	int ip2[32];
	char *ret;
	char *net;
	char *p=".";
		
	if (!(*p_cur) || !p_out1 ||!p_out2)
        return -1;
	ret = strsep(p_cur,":");
	ret = strsep(p_cur," ");
	snprintf(p_out1, length1-1, "%s", ret);
	ret = strsep(p_cur,":");
	return 0;
	
}

//failed -1;success 0
static int formatIPAddr(char **p_cur, char *p_out, int length, int *ip_type)
{
	int i = 0;
	int ip[32];
	char *ret;
	struct sockaddr_in6 sap = {0};
	char tmpaddr[ADDRESS_LEN] = {'\0'};
        RLOGD("formatIPAddr--for--1");
	*ip_type = -1; 

	if (!(*p_cur) || !p_out || length <= 0)
        return -1;

	for(ret = strsep(p_cur,"."); ret != NULL; ret = strsep(p_cur,".")){
		if(strcmp(ret, "") != 0){
			ip[i] = (int)strtoul(ret,NULL,10);
			i++;
		}
	}
        RLOGD("formatIPAddr--for--2");
	if(i < 16){
		//asprintf(p_out,"%d.%d.%d.%d",ip[0],ip[1],ip[2],ip[3]);
		snprintf(p_out, length - 1, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
		*ip_type = TYPE_IPV4;
                RLOGD("formatIPAddr--for--3");
		return 0;
	}
	else{
		//asprintf(p_out,"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",ip[0],ip[1],ip[2],ip[3],ip[4],ip[5],ip[6],ip[7],
		//		ip[8],ip[9],ip[10],ip[11],ip[12],ip[13],ip[14],ip[15]);
		snprintf(tmpaddr, length - 1, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x", ip[0], ip[1], ip[2], ip[3], 
				ip[4], ip[5], ip[6], ip[7], ip[8], ip[9], ip[10], ip[11], ip[12], ip[13], ip[14], ip[15]);

		inet_pton(AF_INET6, tmpaddr, &sap.sin6_addr);
		inet_ntop(AF_INET6, &sap.sin6_addr, p_out, length);
		*ip_type = TYPE_IPV6;
		return 0;
	}
	
}

static int getAddrFromModem(int cid)
{
	ATResponse *p_response = NULL;
	int err;
	ATLine *p_cur;
	char *cmd, *out, *temp_ip, *temp_addr;
	int skip, ip_type, n = 0;
	char local_ip[ADDRESS_LEN] = {'\0'};
	char gateway[ADDRESS_LEN] = {'\0'};
	char tmp_gw[ADDRESS_LEN] = {'\0'};
	char tmp_local_ip[ADDRESS_LEN] = {'\0'};
	char dns1[ADDRESS_LEN] = {'\0'};
	char dns2[ADDRESS_LEN] = {'\0'};
	char pcscf1[ADDRESS_LEN] = {'\0'};
	char pcscf2[ADDRESS_LEN] = {'\0'};
	char tmpdnses[2 * ADDRESS_LEN + 1] = {'\0'}; 
	char tmppcscf[2 * ADDRESS_LEN + 1] = {'\0'}; 
	char dnses[2 * ADDRESS_LEN + 1] = {'\0'};
	char pcscf[2 * ADDRESS_LEN + 1] = {'\0'};


	asprintf(&cmd,"AT+CGCONTRDP=%d", cid);
	err = at_send_command_multiline (cmd, "+CGCONTRDP:", &p_response);
	free(cmd);
	if (err < 0 || p_response->success == 0)
		return -1;
	for (p_cur = p_response->p_intermediates; p_cur != NULL; p_cur = p_cur->p_next){
		n++;
		char *line = p_cur->line;
		err = at_tok_start(&line);
		if (err < 0) return -1;
		err = at_tok_nextint(&line, &skip);
		if (err < 0) return -1;
		err = at_tok_nextint(&line, &skip);
		if (err < 0) return -1;
		err = at_tok_nextstr(&line, &out);
		if (err < 0) return -1;
		err = at_tok_nextstr(&line, &temp_ip);
		if (err < 0) return -1;
		err = formatIPAddr(&temp_ip, &tmp_local_ip, sizeof(tmp_local_ip), &ip_type);
		if (ip_type == TYPE_IPV6){
			err = get_ipv6_local_addr(0, PORT_ECM_KEYNAME, tmp_local_ip, sizeof(tmp_local_ip));
			if (err < 0) return -1;
		}
		err = at_tok_nextstr(&line, &out);  //skip
		if (err < 0) return -1;

		asprintf(&cmd,"AT+CGPDNSADDR=%d", cid);
    		err = at_send_command_singleline (cmd, "+CGPDNSADDR:", &p_response);
		free(cmd);
		if (err < 0 || p_response->success == 0) return -1;
		line = p_response->p_intermediates->line;	
		err = at_tok_start(&line);
		if (err < 0) return -1;
		err = at_tok_nextint(&line,&skip);
		if (err < 0) return -1;
		err = at_tok_nextstr(&line, &temp_ip);
		if (err < 0) return -1;
		err = formatDNSAddr(&temp_ip, &dns1, &dns2, sizeof(dns1));
		if (err < 0) return -1;
		snprintf(&tmpdnses, sizeof(tmpdnses) -1, "%s", dns1);
		err = RIL_ifc_get_default_route(PORT_ECM_KEYNAME, tmp_gw, sizeof(tmp_gw), ip_type);
		if (err < 0) return -1;
		RLOGD("n=%d,tmp_local_ip=%s,tmpdnses=%s,tmp_gw=%s,tmppcscf=%s", n, tmp_local_ip, tmpdnses, tmp_gw, tmppcscf);

		if( n > 1){
			asprintf(&temp_addr, "%s", local_ip);
			snprintf(local_ip, sizeof(local_ip) - 1, "%s %s", temp_addr, tmp_local_ip);
			free(temp_addr);
			if(cid == 3){
				asprintf(&temp_addr, "%s", pcscf);
				snprintf(pcscf, sizeof(pcscf) - 1, "%s %s ", temp_addr, tmppcscf);
			}
			else{
				asprintf(&temp_addr, "%s", dnses);
				snprintf(dnses, sizeof(dnses) - 1, "%s %s", temp_addr, tmpdnses);
			}
			free(temp_addr);
			asprintf(&temp_addr, "%s", gateway);
			snprintf(gateway, sizeof(gateway) - 1, "%s %s", temp_addr, tmp_gw);
			free(temp_addr);
		}else{
			snprintf(local_ip, sizeof(local_ip) - 1, "%s", tmp_local_ip);
			if(cid == 3)
				snprintf(pcscf, sizeof(pcscf) - 1, "%s", tmppcscf);
			else
				snprintf(dnses, sizeof(dnses) -1, "%s", tmpdnses);
			snprintf(gateway, sizeof(gateway) - 1, "%s", tmp_gw);
		}

		//RLOGD("n=%d,local_ip=%s,dnses=%s,gateway=%s,pcscf=%s", n, local_ip, dnses, gateway, pcscf);
	}
	memset(rmnetParam.ip_addr, 0, sizeof(rmnetParam.ip_addr));
	memset(rmnetParam.dns_addr, 0, sizeof(rmnetParam.dns_addr));
	memset(rmnetParam.gateways, 0, sizeof(rmnetParam.gateways));
	memset(rmnetParam.pcscf_addr, 0, sizeof(rmnetParam.pcscf_addr));

	strncpy(rmnetParam.ip_addr, local_ip, strlen(local_ip));
	strncpy(rmnetParam.dns_addr, dnses, strlen(dnses));
	strncpy(rmnetParam.gateways, gateway, strlen(gateway));
	strncpy(rmnetParam.pcscf_addr, pcscf, strlen(pcscf));

	RLOGD("getAddrFromModem:cid=%d,local_ip=%s,dnses=%s,gw=%s,pcscf=%s",cid, 
			rmnetParam.ip_addr, rmnetParam.dns_addr, rmnetParam.gateways,rmnetParam.pcscf_addr);
	
	return 0;

}

/* Query if modem supports PPP. Otherwise, ECM will be used */
static int pppSupported(void)
{
	int err;
	char dial_mode[PROPERTY_VALUE_MAX] = {'\0'};
	int mode = 1;  // 1 ppp, 0 ecm	

	err = property_get("ril.cmiot.dialmode", dial_mode, NULL);
	if(err >= 0 && (strcmp(dial_mode,"ecm") == 0))
		mode = 0;

	RLOGI("dialmode=%d,[1]ppp,[0]ECM",mode);
	
	return mode;
}

static void stopECM(void)
{
    int dhcp_status;
	char *netcfg_down, *dns1, *dns2, *gw;

	asprintf(&netcfg_down,"/system/bin/netcfg %s down",PORT_ECM_KEYNAME);	
	dhcp_status = system(netcfg_down);
	RLOGD("netcfg %s down:status:%d",PORT_ECM_KEYNAME,dhcp_status);

	at_send_command("AT+ZGACT=0,1",NULL);
	at_send_command("AT+CGACT=0,1",NULL);

	system("busybox route del default");  // \C7\E5\B3\FD\CD\F8\B9\D8\D0\C5Ϣ

	asprintf(&dns1,"net.%s.dns1",PORT_ECM_KEYNAME);
	asprintf(&dns2,"net.%s.dns2",PORT_ECM_KEYNAME);
	asprintf(&gw,"net.%s.gw",PORT_ECM_KEYNAME);
	property_set(dns1,"");
	property_set(dns2,"");
	property_set(gw,"");
	free(netcfg_down);
	free(dns1);
	free(dns2);
	free(gw);
}

static int checkECMIsUP(void)
{
	int fd;
	char buffer[20];
	int retry = 3;
	char *operstate_path = NULL;

	asprintf(&operstate_path,"/sys/class/net/%s/operstate",PORT_ECM_KEYNAME);
	
	do{	
		fd = open(operstate_path, O_RDONLY);
		if (fd >= 0) 
		{
			buffer[0] = 0;
			read(fd, buffer, sizeof(buffer));
			close(fd);
			RLOGW("buffer = %s", buffer);
			if(!strncmp(buffer, "up", strlen("up")) || !strncmp(buffer, "unknown", strlen("unknown")))
				break;
		}
		else
		{
			RLOGW("checkECMIsUP can not detect ecm state:%s",strerror(errno));
		}
		sleep(3);
	}while(--retry);

	free(operstate_path);
	
	if(retry <= 0)
		return -1;

	return 0;

}

/* Setup connection using ECM */
static int setupECM(void *data, size_t datalen, RIL_Token t)
{
    int err;
    int dhcp_status;
	char *netcfg_dhcp,*netcfg_up, *gw;
	char *cmd,*line,*apn;
	unsigned int pdp_type_len = 0;
	
	apn = ((const char **)data)[2];
	const char *pdp_type;
	RLOGD("requesting data connection to APN '%s'", apn);

	if (datalen > 6 * sizeof(char *)) {
		pdp_type = ((const char **)data)[6];
	} else {
		pdp_type = "IP";
	}

	pdp_type_len = strlen(pdp_type);
	pdp_type_len = (pdp_type_len > PPP_PDP_TYPE_LEN) ? PPP_PDP_TYPE_LEN : pdp_type_len;
	strncpy(rmnetParam.pdptype,pdp_type,pdp_type_len);	
	at_send_command("AT+CGACT=1,1",NULL);	
	at_send_command("AT+CSTT=\"CMNET\",\"\",\"\"",NULL); 
    at_send_command("AT+CIICR",NULL);	
	sleep(1);
	asprintf(&netcfg_up,"/system/bin/netcfg %s up",PORT_ECM_KEYNAME);	
	asprintf(&netcfg_dhcp,"/system/bin/netcfg %s dhcp",PORT_ECM_KEYNAME);

	dhcp_status = system(netcfg_up);
	RLOGD("netcfg %s up:status:%d",PORT_ECM_KEYNAME,dhcp_status);
	if(dhcp_status < 0)
		goto error;
	dhcp_status = system(netcfg_dhcp);
	RLOGD("netcfg %s dhcp:status:%d",PORT_ECM_KEYNAME,dhcp_status);
	if(dhcp_status < 0)
		goto error;

	//sleep(1);
	err  = checkECMIsUP();
	if(err < 0)
		goto error;

	if(	getAddrFromModem(1) < 0)
		goto error;

	asprintf(&gw, "net.%s.gw", PORT_ECM_KEYNAME);
	property_set(gw, rmnetParam.gateways);
	free(gw);


	RIL_Data_Call_Response_v6 *responses = alloca(sizeof(RIL_Data_Call_Response_v6));

    responses->status = 0;
    responses->suggestedRetryTime = -1;
    responses->cid = 1;
    responses->active = 2;
    responses->type = rmnetParam.pdptype;
    responses->ifname = PORT_ECM_KEYNAME;
    responses->addresses = rmnetParam.ip_addr;
    responses->dnses = rmnetParam.dns_addr;
    responses->gateways = rmnetParam.gateways;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, responses, sizeof(RIL_Data_Call_Response_v6));

	free(netcfg_dhcp);
	free(netcfg_up);
	return 0;

error:
	stopECM();
	free(netcfg_dhcp);
	free(netcfg_up);
	return -1;

}

#if 0
static int waitModemPPPStatus()
{
	//AT+zpppstatus?
	ATResponse *p_response = NULL;
	char *line;
	int err,ret = 0;
	int pppState;

	do{
		RLOGW("waitModemPPPStatus!");
		at_response_free(p_response);
		err = at_send_command_singleline("AT+ZPPPSTATUS?","+ZPPPSTATUS:",&p_response);
		if (err < 0 || p_response->success == 0) 
			continue;
		line = p_response->p_intermediates->line;
		err = at_tok_start(&line);
		if (err < 0) 
			continue;
		err = at_tok_nextint(&line,&pppState);
		if (err < 0) 
			continue;
		if(pppState == 1){
			ret = 1;
			break;
		}
		sleep(1);
	}while(1);

	
	at_response_free(p_response);
	return ret;

}
#endif

static int getModemPPPStatus()
{
	ATResponse *p_response = NULL;
	char *line;
	int err,pppState;
	int ret = 0;

	do{		
		RLOGW("getModemPPPStatus!");
		err = at_send_command_singleline("AT+ZPPPSTATUS?","+ZPPPSTATUS:",&p_response);
		if (err < 0 || p_response->success == 0) 
			break;
		line = p_response->p_intermediates->line;
		err = at_tok_start(&line);
		if (err < 0) 
			break;
		err = at_tok_nextint(&line,&pppState);
		if (err < 0) 
			break;
		if(pppState == 1)
			ret = 1;
		
		RLOGW("getModemPPPStatus:pppState=%d",pppState);
		break;
	}while(1);

	at_response_free(p_response);
	return ret;
}

#define READ_BUF_SIZE 100
static long findPidByName( char* pidName)
{
    DIR *dir_proc;
    struct dirent *next_dir;
    long pid = -1;
    int i=0;
 
 	/* Open the /proc directory. */
    dir_proc = opendir("/proc");
    if (!dir_proc)
        RLOGD("Cannot open /proc");
	
    while ((next_dir = readdir(dir_proc)) != NULL) {
        FILE *status;
        char filename[100];
        char buffer[READ_BUF_SIZE];
        char name[READ_BUF_SIZE];
 
        /* Must skip ".." since that is outside /proc */
        if (strcmp(next_dir->d_name, "..") == 0)
            continue;
 
        /* If it isn't a number, we don't want it */
        if (!isdigit(*next_dir->d_name))
            continue;
        
        sprintf(filename, "/proc/%s/cmdline", next_dir->d_name);
        if (! (status = fopen(filename, "r")) ) {
            continue;
        }
        if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
            fclose(status);
            continue;
        }
        fclose(status);
 
        /* Buffer should contain a string like "Name:   binary_name" */
        sscanf(buffer, "%s", name);
		//RLOGD("%s:%s",filename,name);
        if (strcmp(name, pidName) == 0) {
            //pidList=realloc(pidList, sizeof(long) * (i+2));
            pid = strtol(next_dir->d_name, NULL, 0);
			
			RLOGD("find %s 's pid is %ld",pidName,pid);
			break;
        }
    }

    return pid;
}

static void killPPPd(void)
{
	long pppd_pid;	
    int status = -1;

	pppd_pid = findPidByName("/system/bin/pppd");
	if(pppd_pid> 0)
	{	
		do 
		{
			status = kill(pppd_pid,SIGKILL);	
			RLOGD("killPPPd status:%d",status);
			sleep(1);
		}while(status < 0);
		//wait(&status);
	}
}

static void stopPPPd(void)
{
	int fd;
	
	RLOGD("stopPPPd!");
	
	at_send_command("AT+CMPDPDEACT=1", NULL);

	at_send_command("ATH", NULL);	//\CC\ED\BCӹر\D5

	sleep(1);
	if(PPPD_START(sMdmInfo) != PPP_CLOSED)
	{	
		RLOGD("stopping pppd");
		property_set("net.gprs.enable", "0");
		do{			
			fd = open("/sys/class/net/ppp0/operstate", O_RDONLY);
			if (fd < 0)
				break;
			else
				RLOGD("can't kill pppd!");

			close(fd);
			sleep(1);
		} while (1);
		
		PPPD_START(sMdmInfo) = PPP_CLOSED;
	}
	
	property_set("net.ppp0.local-ip","");
	property_set("net.ppp0.dns1","");
	property_set("net.ppp0.dns2","");
	property_set("net.ppp0.remote-ip","");
	property_set("net.gprs.ppp-exit","");
	
}

static void onDeactiveDataCallList(void)
{
    RIL_Data_Call_Response_v6 *responses = alloca(sizeof(RIL_Data_Call_Response_v6));

	if(USE_PPP(sMdmInfo)){
		stopPPPd();
	}else{
		stopECM();
	}
			
    responses->status = -1;
    responses->suggestedRetryTime = -1;
    responses->cid = 1;
    responses->active = 0;
    responses->type = "";
    responses->ifname = PORT_ECM_KEYNAME;
    responses->addresses = "";
    responses->dnses = "";
    responses->gateways = "";

    RIL_onUnsolicitedResponse(RIL_UNSOL_DATA_CALL_LIST_CHANGED,responses,sizeof(RIL_Data_Call_Response_v6));

    RLOGI("onDeactiveDataCallList!");
}

static int checkPPPIsUPByPPPLoop(int retry)
{
	char exit_code[PROPERTY_VALUE_MAX] = {'\0'};
	int fd;
	int err = 0;
	char buffer[20];
	char local_ip[PROPERTY_VALUE_MAX] = {'\0'};

	do{	
		err = property_get("net.gprs.ppp-exit", exit_code, NULL);
		if((err >= 0) && (strcmp(exit_code, "") != 0)){ 
			RLOGW("PPPLoop PPPd exit with code %s", exit_code);
			retry = 0;
			break;
		}
		
		fd = open("/sys/class/net/ppp0/operstate", O_RDONLY);
		if (fd >= 0) {
			buffer[0] = 0;
			read(fd, buffer, sizeof(buffer));
			close(fd);
			RLOGW("checkPPPIsUPByPPPLoop:buffer = %s", buffer);
			if(!strncmp(buffer, "up", strlen("up")) || !strncmp(buffer, "unknown", strlen("unknown"))){
				err = property_get("net.ppp0.local-ip", local_ip, NULL);
				if((err >= 0) && (strcmp(local_ip, "") != 0)){ 
					break;
				}else{
					RLOGW("ppp0 link is up,but no local_ip is assigned!");
				}
			}
		}else{
			RLOGW("checkPPPIsUPByPPPLoop can not detect PPP state:%s",strerror(errno));
		}
		
		sleep(1);
	}while(retry--);

	if(retry <= 0)
		return -1;

	return 0;
}

static int checkPPPIsUP(int retry)
{
	char exit_code[PROPERTY_VALUE_MAX] = {'\0'};
	int fd;
	int err = 0;
	char buffer[20];
	char local_ip[PROPERTY_VALUE_MAX] = {'\0'};
	
	do{	
		err = property_get("net.gprs.ppp-exit", exit_code, NULL);
		if((err >= 0) && (strcmp(exit_code, "") != 0)){ 
			RLOGW("PPPd exit with code %s", exit_code);
			retry = 0;
			break;
		}
		
		fd = open("/sys/class/net/ppp0/operstate", O_RDONLY);
		if (fd >= 0) 
		{
			buffer[0] = 0;
			read(fd, buffer, sizeof(buffer));
			close(fd);
			RLOGW("checkPPPIsUP:buffer = %s", buffer);
			if(!strncmp(buffer, "up", strlen("up")) || !strncmp(buffer, "unknown", strlen("unknown")))
			{
				err = property_get("net.ppp0.local-ip", local_ip, NULL);
				if((err >= 0) && (strcmp(local_ip, "") != 0)){ 
					break;
				}else{
					RLOGW("ppp0 link is up,but no local_ip is assigned!");
				}
			}
		}else{
			RLOGW("checkPPPIsUP can not detect PPP state:%s",strerror(errno));
		}
		sleep(2);
	}while(retry--);

	if(retry <= 0){	
		PPPD_START(sMdmInfo) = PPP_CLOSING;
		return -1;
	}

	return 0;
}

static int setupPPP(void *data, size_t datalen, RIL_Token t)
{
	const char *apn;
	char *cmd,*line;
	int err,cgatt_state;
	unsigned int pdp_type_len = 0;
		
	ATResponse *p_response = NULL;

	RIL_Data_Call_Response_v6 *responses = alloca(sizeof(RIL_Data_Call_Response_v6));

	apn = ((const char **)data)[2];

	int retry = 10;
	const char *pdp_type;

	if (datalen > 6 * sizeof(char *)) {
		pdp_type = ((const char **)data)[6];
	} else {
		pdp_type = "IP";
	}
	
	pdp_type_len = strlen(pdp_type);
	pdp_type_len = (pdp_type_len > PPP_PDP_TYPE_LEN) ? PPP_PDP_TYPE_LEN : pdp_type_len;
	strncpy(rmnetParam.pdptype,pdp_type,pdp_type_len);
	
	stopPPPd();
	//sleep(1);
	at_send_command("AT+CGACT=0,1",NULL);	
	
	if(	USE_VOLTE(sMdmInfo) == 1)
		asprintf(&cmd, "AT+ZGDCONT=1,\"%s\",\"%s\"", rmnetParam.pdptype, apn);
	else
		asprintf(&cmd, "AT+CGDCONT=1,\"%s\",\"%s\"", rmnetParam.pdptype, apn);
	err = at_send_command(cmd, NULL);
	free(cmd);
	/*sleep(1);
	err = at_send_command_singleline("AT+CGATT?","+CGATT:",&p_response);
	if (err < 0 || p_response->success == 0)
		goto error;
	line = p_response->p_intermediates->line;
	err = at_tok_start(&line);
	if (err < 0) 
		goto error;
	err = at_tok_nextint(&line,&cgatt_state);
	if(err < 0 || cgatt_state == 0){
		at_send_command("AT+CGATT=1",NULL);
		goto error;
	}*/
	
	RLOGE("setupPPP: requesting data connection to APN '%s',%s", apn,rmnetParam.pdptype);

   	err = property_set("net.gprs.enable", "1");
   	if (err < 0) {
    	RLOGW("Can not start PPPd");
       	goto error;
	}
	
	PPPD_START(sMdmInfo) = PPP_STARTING;
	RLOGD("PPPD starting!");
#if 0	
	do {
		if(ppp_fd < 0){
			ppp_fd = open(PPP_TTY_PATH, O_RDWR);
			//ppp_fd = open("/dev/ttyUSB0", O_RDWR);
			RLOGD("ppp_fd:%d",ppp_fd);
		}

		if(ppp_fd >= 0){
			struct termios  ios;
			tcgetattr( ppp_fd, &ios );
			ios.c_lflag = 0;  /* disable ECHO, ICANON, etc... */
			tcsetattr( ppp_fd, TCSANOW, &ios );
			tcflush(ppp_fd,TCIOFLUSH);
			err = writelineEx(ppp_fd, "ATD*99***1#");
			if(err < 0){
				RLOGE("send dial command failed!");
			}
				RLOGI("send dial command ATD*99***1#!");
			break;
		}else{	 
			close(ppp_fd);
			ppp_fd = -1;
			RLOGW("retry %d	after 3 seconds", retry);
			sleep(3);
		}
	} while (--retry);
	if((ppp_fd < 0)&&(retry == 0)){
		RLOGE("open modem fail!");
		goto error;
	}
#endif

	sleep(5);
	err = checkPPPIsUP(20);
	if(err < 0 )
		goto error;
	
	PPPD_START(sMdmInfo) = PPP_STARTED;
	if(getAddressInfo() < 0)
		goto error;

	responses->status = 0;
	responses->suggestedRetryTime = -1;
	responses->cid = 1;
	responses->active = 2;
	responses->type = rmnetParam.pdptype;
	responses->ifname = "ppp0";
	responses->addresses = rmnetParam.ip_addr;
	responses->dnses = rmnetParam.dns_addr;
	responses->gateways = rmnetParam.gateways;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, responses, sizeof(RIL_Data_Call_Response_v6));
	at_response_free(p_response);
	return 0;
	
error:
	
	stopPPPd();
	at_response_free(p_response);
	
	return -1;
}


static void requestSetupDataCall(void *data, size_t datalen, RIL_Token t)
{
	RLOGI("requestSetupDataCall");
	memset(&rmnetParam,0,sizeof(rmnetParam));
	
	if(USE_PPP(sMdmInfo)){
        RLOGI("requestSetupDataCall-PPP");
		if(setupPPP(data,datalen,t) < 0)
			goto error;
	}
	else{
        RLOGI("requestSetupDataCall-ECM");
		if(setupECM(data,datalen,t) < 0)
	    	goto error;
	}
	
	return;
error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	return;

}

static void requestDeactivateDataCall(void *data, size_t datalen, RIL_Token t)
{
	if(USE_PPP(sMdmInfo)){
		stopPPPd();
	}else{
		stopECM();
	}

	RLOGD("requestDeactivateDataCall");
	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestLastDataCallFailCause(RIL_Token t)
{

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &s_lastPdpFailCause, sizeof(int));
}

static unsigned int hex2int(char dig)
{
	if (dig >= '0' && dig <= '9')
		return dig - '0';
	if (dig >= 'a' && dig <= 'f')
		return dig - 'a' + 10;
	if (dig >= 'A' && dig <= 'F')
		return dig - 'A' + 10;
	return 0;
}

static void hexStringToBytes(unsigned char* b, const char* str)
{
	int len = strlen(str);
	len >>= 1;
	while (len--) {
		*b++ = (hex2int(str[0]) << 4) | hex2int(str[1]);
		str += 2;
	};
}

static void  requestSIM_IO(void *data, size_t datalen __unused, RIL_Token t)
{
	ATResponse *p_response = NULL;
	RIL_SIM_IO_Response sr;
	int err;
	char *cmd = NULL;
	RIL_SIM_IO_v6 *p_args;
	char *line;
	memset(&sr, 0, sizeof(sr));
	p_args = (RIL_SIM_IO_v6 *)data;

	if (!p_args) 
		goto error;

	if (p_args->path != NULL && (strlen(p_args->path) & 3) != 0)
		goto error;
        

	if (p_args->data == NULL) {
		if (p_args->path == NULL) {
			asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d", p_args->command, p_args->fileid,
					p_args->p1, p_args->p2, p_args->p3);
		} else {
			asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d,,\"%s\"", p_args->command,
					p_args->fileid, p_args->p1, p_args->p2, p_args->p3, p_args->path);
		}
	} else {
		if (p_args->path == NULL) {
			asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d,%s", p_args->command, p_args->fileid,
					p_args->p1, p_args->p2, p_args->p3, p_args->data);
		} else {
			asprintf(&cmd, "AT+CRSM=%d,%d,%d,%d,%d,%s,\"%s\"", p_args->command, p_args->fileid,
					p_args->p1, p_args->p2, p_args->p3, p_args->data, p_args->path);
		}
	}  

	//err = at_send_command_singleline("AT+CRSM=179,20282,2,4,28,,\"7F105F3A\"", "+CRSM:", &p_response);
    err = at_send_command_singleline(cmd, "+CRSM:", &p_response);
	if (err < 0 || p_response->success == 0) {
		goto error;
	}
	line = p_response->p_intermediates->line;
	err = at_tok_start(&line);
	if (err < 0) goto error;
	err = at_tok_nextint(&line, &(sr.sw1));
	if (err < 0) goto error;
	err = at_tok_nextint(&line, &(sr.sw2));
	if (err < 0) goto error;
	if (at_tok_hasmore(&line)) {
		err = at_tok_nextstr(&line, &(sr.simResponse));
		if (err < 0) goto error;
	}

	if (sr.simResponse != NULL) {
		struct ts_51011_921_resp ts_data;
		if (fcp_to_ts_51011(sr.simResponse, strlen(sr.simResponse), &ts_data) == 0) {
			err = binaryToString(&ts_data, sizeof(ts_data), sr.simResponse);
			RLOGE("requestSIM_IO() simResponse-->:%s",sr.simResponse);
		}
	}
#if 1
		/* If dealing with a USIM card ... */
		if (p_args->command == 0xC0 &&
			sr.simResponse != NULL &&
			sr.simResponse[0] == '6' &&
			sr.simResponse[1] == '2' &&
			sr.simResponse[6] == '0' &&
			sr.simResponse[7] == '5' &&
			strnlen(sr.simResponse, 40) <= 30 &&
			getUICCType() != UICC_TYPE_SIM) {
			
			/* Convert it to a format Android understands */
			
			unsigned char buf[15];
			unsigned int val;
			
			// Convert to bytes...
			hexStringToBytes(buf,sr.simResponse);
			
			// Reformat response...
			sr.simResponse[0x00] = '0';
			sr.simResponse[0x01] = '0';
			sr.simResponse[0x02] = '0';
			sr.simResponse[0x03] = '0';
			
			val = buf[8] * (buf[7] + (buf[6]<<8U));
			sprintf(&sr.simResponse[0x04],"%04x",val & 0xFFFFU);
			
			sr.simResponse[0x08] = sr.simResponse[0x16];
			sr.simResponse[0x09] = sr.simResponse[0x17];
			sr.simResponse[0x0A] = sr.simResponse[0x18];
			sr.simResponse[0x0B] = sr.simResponse[0x19];
			sr.simResponse[0x0D] = '4';
			sr.simResponse[0x1A] = '0';
			sr.simResponse[0x1B] = '1';
			sr.simResponse[0x1C] = sr.simResponse[0x0E];
			sr.simResponse[0x1D] = sr.simResponse[0x0F];
			sr.simResponse[0x1E] = 0;
		}
#endif

	RIL_onRequestComplete(t, RIL_E_SUCCESS, &sr, sizeof(sr));
	at_response_free(p_response);
	free(cmd);
	return;
error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
	free(cmd);
}

static void  requestEnterSimPin(int request, void*  data, size_t  datalen, RIL_Token  t)
{
    ATResponse   *p_response = NULL;
    int           err;
    char*         cmd = NULL;
    const char**  strings = (const char**)data;;
    int num_retries = 1;
	
    if ( datalen == sizeof(char*) ) {
        asprintf(&cmd, "AT+CPIN=%s", strings[0]);
    } else if ( datalen == 2*sizeof(char*) ) {
        asprintf(&cmd, "AT+CPIN=%s,%s", strings[0], strings[1]);
    } else
        goto error;

    err = at_send_command_singleline(cmd, "+CPIN:", &p_response);
    free(cmd);
	if(err < 0)
		goto error;
	if(p_response->success == 0){
		RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, &num_retries, sizeof(int*));
	}else{
		RIL_onRequestComplete(t, RIL_E_SUCCESS, &num_retries, sizeof(int*));
		setRadioState(RADIO_STATE_SIM_NOT_READY);
	}

	at_response_free(p_response);
 	return;	
error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE,  &num_retries, sizeof(int*));
    at_response_free(p_response);
}

static void  requestChangePassword(int request, void*  data, size_t  datalen, RIL_Token  t)
{
	ATResponse *p_response = NULL;
	int err = 0;
	char *cmd = NULL;

	char *string = NULL;
	char *old_password = NULL;
	char *new_password = NULL;

	if (datalen != 3 * sizeof(char *))
		goto error;

	old_password = ((char **)data)[1];
	new_password = ((char **)data)[2];

	if(request == RIL_REQUEST_CHANGE_SIM_PIN)
		asprintf(&string,"SC");
	else if(request == RIL_REQUEST_CHANGE_SIM_PIN2)
		asprintf(&string,"P2");
		
	asprintf(&cmd,"AT+CPWD=\"%s\",\"%s\",\"%s\"", string, old_password, new_password);
	err = at_send_command(cmd,&p_response);
	free(cmd);
	free(string);
	if (err < 0 || p_response->success == 0)
		goto error;

	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	at_response_free(p_response);
	return;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}

static void requestChangeBarringPassword(void *data, size_t datalen, RIL_Token t)
{
	ATResponse *p_response = NULL;
	int err = 0;
	char *cmd = NULL;

    char *string = NULL;
    char *old_password = NULL;
    char *new_password = NULL;

    assert (datalen >=  (3 * sizeof(char **)));

    string     = ((char **)data)[0];
    old_password = ((char **)data)[1];
    new_password = ((char **)data)[2];

	asprintf(&cmd,"AT+CPWD=\"%s\",\"%s\",\"%s\"", string, old_password, new_password);
    err = at_send_command(cmd,&p_response);
	free(cmd);
    if (err < 0 || p_response->success == 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}

static void  requestSendUSSD(void *data, size_t datalen __unused, RIL_Token t)
{
    const char *ussdRequest;

    ussdRequest = (char *)(data);

    RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);

}

static void requestCancelUSSD(RIL_Token t)
{
	ATResponse   *p_response = NULL;
    int           err;
	
	err = at_send_command_numeric("AT+CUSD=2", &p_response);
	
	if (err < 0 || p_response->success == 0) {
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	} else {
		RIL_onRequestComplete(t, RIL_E_SUCCESS,p_response->p_intermediates->line, sizeof(char *));
	}
	at_response_free(p_response);
}

static void requestQueryCallWaiting(void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err;
    char *line;
    int aclass;
    int response[2];
	char * cmd = NULL;
	
    aclass = ((int *)data)[0];

	asprintf(&cmd,"AT+CCWA=1,2,%d",aclass);
    err = at_send_command_singleline(cmd, "+CCWA:", &p_response);
	free(cmd);
	
	//asprintf(&cmd,"AT+ZCWA=1,2,%d",aclass);
    //err = at_send_command_singleline(cmd, "+ZCWA:", &p_response);
	//free(cmd);
	
    if (err < 0 || p_response->success == 0)
        goto error;

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(response[0]));
    if (err < 0) goto error;

    if (at_tok_hasmore(&line)) {
        err = at_tok_nextint(&line, &(response[1]));
        if (err < 0) goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestSetCallWaiting(void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err;
    int enabled, aclass;
	char * cmd = NULL;

	if((datalen<2)||(data==NULL)) goto error;

    enabled = ((int *)data)[0];
    aclass = ((int *)data)[1];

	asprintf(&cmd,"AT+CCWA=0,%d,%d",enabled,aclass);
	//asprintf(&cmd,"AT+ZCWA=0,%d,%d",enabled,aclass);
    err = at_send_command(cmd,&p_response);
	free(cmd);
	
    if (err < 0 || p_response->success == 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestQueryCallForwardStatus(RIL_Token t)
{
    int err = 0;
    int i = 0;
    int n = 0;
    int tmp = 0;
    ATResponse *p_response = NULL;
    ATLine *p_cur;
    RIL_CallForwardInfo **responses = NULL;
	
	//if(TECH(sMdmInfo)  == MDM_LTE)
		err = at_send_command_multiline("AT+ZCFC=0,2", "+ZCFC:", &p_response);
	//else 
		//err = at_send_command_multiline("AT+CCFC=0,2", "+CCFC:", &p_response);

    if (err < 0 || p_response->success == 0)
        goto error;

    for (p_cur = p_response->p_intermediates; p_cur != NULL; p_cur = p_cur->p_next)
        n++;

    responses = (RIL_CallForwardInfo **) alloca(n * sizeof(RIL_CallForwardInfo *));

    for (i = 0; i < n; i++) {
        responses[i] = (RIL_CallForwardInfo*)alloca(sizeof(RIL_CallForwardInfo));
        responses[i]->status = 0;
        responses[i]->reason = 0;
        responses[i]->serviceClass = 0;
        responses[i]->toa = 0;
        responses[i]->number = (char*)"";
        responses[i]->timeSeconds = 0;
    }

    for (i = 0,p_cur = p_response->p_intermediates; p_cur != NULL; p_cur = p_cur->p_next, i++) {
        char *line = p_cur->line;

        err = at_tok_start(&line);
        if (err < 0) goto error;

        err = at_tok_nextint(&line, &(responses[i]->status));
        if (err < 0) goto error;

        err = at_tok_nextint(&line, &(responses[i]->serviceClass));
        if (err < 0) goto error;

        if(!at_tok_hasmore(&line)) continue;

        err = at_tok_nextstr(&line, &(responses[i]->number));
        if (err < 0) goto error;

        if(!at_tok_hasmore(&line)) continue;

        err = at_tok_nextint(&line, &(responses[i]->toa));
        if (err < 0) goto error;

        if(!at_tok_hasmore(&line)) continue;
        at_tok_nextint(&line,&tmp);

        if(!at_tok_hasmore(&line)) continue;
        at_tok_nextint(&line,&tmp);

        if(!at_tok_hasmore(&line)) continue;
        err = at_tok_nextint(&line, &(responses[i]->timeSeconds));
        if (err < 0) goto error;
    }

    RIL_onRequestComplete(t, RIL_E_SUCCESS, responses, sizeof(RIL_CallForwardInfo **));
    at_response_free(p_response);

    return;
error:
    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetCallForward(void *data, RIL_Token t)
{
    ATResponse *p_response = NULL;
	int err = 0;
    RIL_CallForwardInfo *info = NULL;
	char *cmd = NULL;
		
    info = ((RIL_CallForwardInfo *) data);

    if(data == NULL)
        goto error;

	//if(TECH(sMdmInfo)  == MDM_LTE)
		asprintf(&cmd,"AT+ZCFC=%d,%d,\"%s\",%d",info->reason, info->status,info->number,info->toa);
	//else
		//asprintf(&cmd,"AT+CCFC=%d,%d,\"%s\",%d",info->reason, info->status,info->number,info->toa);
    err = at_send_command(cmd,&p_response);
	free(cmd);
    if (err < 0 || p_response->success == 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}

static void requestGetCLIR(void *data, size_t datalen, RIL_Token t)
{
    /* Even though data and datalen must be NULL and 0 respectively this
     * condition is not verified
     */
    ATResponse *p_response = NULL;
    int response[2];
    char *line = NULL;
    int err = 0;

    err = at_send_command_singleline("AT+CLIR?", "+CLIR:", &p_response);
    if (err < 0 || p_response->success == 0)
        goto error;

    line = p_response->p_intermediates->line;
    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(response[0]));
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &(response[1]));
    if (err < 0) goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, response, sizeof(response));
    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestSetCLIR(void *data, size_t datalen, RIL_Token t)
{
	ATResponse *p_response = NULL;
	int err = 0;
	char *cmd = NULL;

	asprintf(&cmd,"AT+CLIR=%d", ((int *)data)[0]);
    err = at_send_command(cmd,&p_response);
	free(cmd);
    if (err < 0 || p_response->success == 0)
        RIL_onRequestComplete(t, RIL_E_PASSWORD_INCORRECT, NULL, 0);
    else
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);

	at_response_free(p_response);
}

static void requestQueryCLIP(RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err;
    char *line;
    int response;

    err = at_send_command_singleline("AT+CLIP?","+CLIP:",&p_response);
    if (err < 0 || p_response->success == 0)
        goto error;

    line = p_response->p_intermediates->line;
    err = at_tok_start(&line);
    if (err < 0) goto error;

    /* The first number is discarded */
    err = at_tok_nextint(&line, &response);
    if (err < 0) goto error;

    err = at_tok_nextint(&line, &response);
    if (err < 0) goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(int));
    at_response_free(p_response);
    return;

error:
    at_response_free(p_response);
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestResetRadio(RIL_Token t)
{
    int err = 0;

    /* Reset MS */
    err = at_send_command("AT+CFUN=4",NULL);

    /* Go online */
    err = at_send_command("AT+CFUN=1",NULL);
    if (err < 0)
        goto error;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestSetSuppSVCNotification(void *data, size_t datalen, RIL_Token t)
{
	ATResponse *p_response = NULL;
	int err = 0;
    int enabled = 0;
	char *cmd = NULL;
    enabled = ((int *)data)[0];

	asprintf(&cmd,"AT+CSSN=%d,%d", enabled, enabled);
    err = at_send_command(cmd,&p_response);
	free(cmd);
    if (err < 0 || p_response->success == 0)
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    else
        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);

	at_response_free(p_response);
}

static void requestSetLocationUpdates(void *data, size_t datalen, RIL_Token t)
{
	int err = 0;
    int updates = 0;
	
    updates = ((int *)data)[0] == 1? 2 : 1;
	if(updates == 1){
		err = at_send_command("AT+CREG=1",NULL);
    	if (err < 0)
        	goto error;
		err = at_send_command("AT+CGREG=1",NULL);
    	if (err < 0)
        	goto error;
		err = at_send_command("AT+CEREG=1",NULL);
    	if (err < 0)
        	goto error;
	}else if(updates == 2){
		err = at_send_command("AT+CREG=2",NULL);
    	if (err < 0)
        	goto error;
		err = at_send_command("AT+CGREG=2",NULL);
    	if (err < 0)
        	goto error;
		err = at_send_command("AT+CEREG=2",NULL);
    	if (err < 0)
        	goto error;
	}else
		goto error;
	
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
}

static void requestLastCallFailCause(RIL_Token t)
{
    ATResponse *p_response = NULL;
    int err = 0;
    int error_type = -1;
	int error_info = -1;
    char *line = NULL;

    err = at_send_command_singleline("AT+CEER", "+CEER:", &p_response);
    if (err < 0 || p_response->success == 0)
        goto error;

	/*
	CALL_FAIL_NORMAL = 16,
    CALL_FAIL_BUSY = 17,
    CALL_FAIL_CONGESTION = 34,
    CALL_FAIL_ACM_LIMIT_EXCEEDED = 68,
    CALL_FAIL_CALL_BARRED = 240,
    CALL_FAIL_FDN_BLOCKED = 241,
    CALL_FAIL_ERROR_UNSPECIFIED = 0xffff
	*/
		
    line = p_response->p_intermediates->line;
    err = at_tok_start(&line);
    if(err < 0) goto error;

    err = at_tok_nextint(&line, &error_type);
    if(err < 0) goto error;

    if (at_tok_hasmore(&line)) {
		err = at_tok_nextint(&line, &error_info);
		if(err < 0) goto error;
	}

	RLOGD("Textual last call fail cause: %d [%d]", error_type, error_info);

	if(error_type == 0 && error_info != -1)
		cc_lastCallFailCause = error_info;
	
error:
	/* Send the last call fail cause */
    RIL_onRequestComplete(t, RIL_E_SUCCESS, &cc_lastCallFailCause, sizeof(int));
    at_response_free(p_response);
    return;
}

// TODO: Use all radio types
static int techFromModemType(int mdmtype)
{
    int ret = -1;
    switch (1 << mdmtype) {
        case MDM_CDMA:
            ret = RADIO_TECH_1xRTT;
            break;
        case MDM_EVDO:
            ret = RADIO_TECH_EVDO_A;
            break;
        case MDM_GSM:
            ret = RADIO_TECH_GPRS;
            break;
        case MDM_WCDMA:
            ret = RADIO_TECH_HSPA;
            break;
		case MDM_TD_SCDMA:
		ret  = RADIO_TECH_TD_SCDMA;
		break;
        case MDM_LTE:
            ret = RADIO_TECH_LTE;
            break;
    }
    return ret;
}

/* static void requestGetCellInfoList(void *data __unused, size_t datalen __unused, RIL_Token t)
{
    uint64_t curTime = ril_nano_time();
    RIL_CellInfo ci[1] =
    {
        { // ci[0]
            1, // cellInfoType
            1, // registered
            2,
            curTime - 1000, // Fake some time in the past
            { // union CellInfo
                {  // RIL_CellInfoGsm gsm
                    {  // gsm.cellIdneityGsm
                        s_mcc, // mcc
                        s_mnc, // mnc
                        s_lac, // lac
                        s_cid, // cid
                        0  // psc
                    },
                    {  // gsm.signalStrengthGsm
                        10, // signalStrength
                        0  // bitErrorRate
                    }
                }
            }
        }
    };

    RIL_onRequestComplete(t, RIL_E_SUCCESS, ci, sizeof(ci));
}

static void requestSetCellInfoListRate(void *data, size_t datalen, RIL_Token t)
{
    // For now we'll save the rate but no RIL_UNSOL_CELL_INFO_LIST messages
    // will be sent.
    assert (datalen == sizeof(int));
    s_cell_info_rate_ms = ((int *)data)[0];

    RLOGI("requestSetCellInfoListRate:%d",s_cell_info_rate_ms);	
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}
*/

static void requestNeighboringCellIds(void * data, size_t datalen, RIL_Token t)
{
	RIL_NeighboringCell **pp_cellIds;
    RIL_NeighboringCell *p_cellIds;

    pp_cellIds = (RIL_NeighboringCell **)alloca(sizeof(RIL_NeighboringCell *));
    p_cellIds = (RIL_NeighboringCell *)alloca(sizeof(RIL_NeighboringCell));
    pp_cellIds[0]=p_cellIds;

}

#if 0
static void requestGetHardwareConfig(void *data, size_t datalen, RIL_Token t)
{
   // TODO - hook this up with real query/info from radio.

   RIL_HardwareConfig hwCfg;

   RIL_UNUSED_PARM(data);
   RIL_UNUSED_PARM(datalen);

   hwCfg.type = -1;

   RIL_onRequestComplete(t, RIL_E_SUCCESS, &hwCfg, sizeof(hwCfg));
}
#endif

static void requestGetRadioCapability(void *data, size_t datalen, RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}


static void requestSetRadioCapability(void *data, size_t datalen, RIL_Token t)
{
	RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}


static void requestGetIMSI(RIL_Token t)
{
    	ATResponse   *p_response = NULL;
	 int err;
	char *line,*imsi;

	// err = at_send_command_numeric("AT+CIMI", &p_response);
	
	/* IMSI */
	err = at_send_command_numeric("AT+CIMI", &p_response);

	if (err < 0 || p_response->success == 0) 
		goto error;
			
	RIL_onRequestComplete(t, RIL_E_SUCCESS, p_response->p_intermediates->line, sizeof(char *));


	at_response_free(p_response);
	return;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);
}

static void requestGetIMEI(RIL_Token t)
{
    	ATResponse   *p_response = NULL;
	 int err;
	char *line,*imei;

	err = at_send_command_numeric("AT+GSN=1", &p_response);
	

       if (err < 0 || p_response->success == 0) 
		goto error;
			
	RIL_onRequestComplete(t, RIL_E_SUCCESS, p_response->p_intermediates->line, sizeof(char *));


	at_response_free(p_response);	

	return;

error:
	RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	at_response_free(p_response);

}

static void requestScreenState(void *data, size_t datalen, RIL_Token token)
{

	static int screenState = 0;
	screenState  = ((int *) data)[0];

	RLOGD("requestScreenState");
	if (getRadioState() != RADIO_STATE_UNAVAILABLE)
	{
		if (!screenState)//notify CP that AP will sleep
		{
			RLOGD("tim:AT+ZPOWERIND begin ");
			at_send_command("AT+ZPOWERIND", NULL);
		}
		else
		{
			/*** Note:  when screen on, update +CREG state***/
			RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_VOICE_NETWORK_STATE_CHANGED, NULL, 0);
		}
	}
	

	RIL_onRequestComplete(token, RIL_E_SUCCESS, NULL, 0);

}

static void requestGetSimStatus(RIL_Token t)
{
	RIL_CardStatus_v6 *p_card_status = NULL;
	char *p_buffer;
	    int buffer_size;

	RLOGI("requestGetSimStatus");
    int result = getCardStatus(&p_card_status);
    if (result == RIL_E_SUCCESS) {
        p_buffer = (char *)p_card_status;
        buffer_size = sizeof(*p_card_status);
    } else {
    	result = RIL_E_GENERIC_FAILURE;
        p_buffer = NULL;
        buffer_size = 0;
    }
	RLOGD("requestGetSimStatus:%d,%s,%d",result,p_buffer,buffer_size);
    RIL_onRequestComplete(t, result, p_buffer, buffer_size);
    if (p_card_status != NULL)
		freeCardStatus(p_card_status);
}

static void requestCdmaSubscriptionSource(void * data, size_t datalen, RIL_Token t)
{

    /*
    * "data" is int *
    * ((int *)data)[0] is == 0 from RUIM/SIM (default)
    * ((int *)data)[0] is == 1 from NV
    */

    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestReportStkServiceIsRunning(RIL_Token t)
{
    RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}


static void requestOemHookStrings(void * data, size_t datalen, RIL_Token t)
{
	int i;
	const char ** cur;
	int err;
	ATResponse *p_response = NULL;
	RLOGD("got OEM_HOOK_STRINGS: 0x%8p %lu", data, (long)datalen);
	for (i = (datalen / sizeof (char *)), cur = (const char **)data ;i > 0 ; cur++, i --) {
		RLOGD("> '%s'", *cur);
	}

	err = at_send_command_multiline( ((const char **)data)[0], "\0", &p_response);
	if(err < 0 || p_response == NULL || p_response->success == 0)
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	else{
		char *lines[100] ;
		size_t i = 0;
		ATLine	*p_intermediates = p_response->p_intermediates;
		while (p_intermediates && (i < ((sizeof(lines) / sizeof(lines[0])) - 1))) {
			if (p_intermediates->line)
				lines[i++] = p_intermediates->line;
			p_intermediates = p_intermediates->p_next;
		}
		if (p_response->finalResponse)
			lines[i++] = p_response->finalResponse;
		RIL_onRequestComplete(t, RIL_E_SUCCESS, lines, i*sizeof(char *));
	}
}

static void requestBasebandVersion(RIL_Token t)
{
    int err;
    ATResponse *p_response = NULL;
    char *line;

    err = at_send_command_singleline("AT+GMR", "\0", &p_response);

    if(err < 0 || p_response == NULL || p_response->success == 0) {
        RLOGD("Error reading Base Band Version");
        RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
        return;
    }
 
    line = p_response->p_intermediates->line;

    RIL_onRequestComplete(t, RIL_E_SUCCESS, line, sizeof(char *));

    at_response_free(p_response);
}

static void requestQueryFacilityLock(void *data, size_t datalen, RIL_Token t)
{

}

static void requestSetFacilityLock(void *data, size_t datalen, RIL_Token t)
{

}

/**
 * RIL_REQUEST_DEVICE_IDENTITY
 *
 * Request the device ESN / MEID / IMEI / IMEISV.
 *
 */
static void requestDeviceIdentity(RIL_Token t)
{
    ATResponse *p_response = NULL;
    char* response[4];
    int err;
	char *line,*imei;

    /* IMEI */
    	err = at_send_command_singleline("AT+GSN=1", "+GSN:", &p_response);

    if (err < 0 || p_response->success == 0)
        goto error;

	line =  p_response->p_intermediates->line;
	err = at_tok_start(&line);
	if(err<0) goto error;

	err = at_tok_nextstr(&line,&imei);
	if(err<0) goto error;
	
    response[0] = imei;

    /* IMEISVN */
    response[1] = (char*) "01";

    /* CDMA not supported */
    response[2] = (char*) "";
    response[3] = (char*) "";

    RIL_onRequestComplete(t, RIL_E_SUCCESS, &response, sizeof(response));

    at_response_free(p_response);
    return;

error:
    RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
    at_response_free(p_response);
}

static void requestSeparateConnection(void *data, size_t datalen __unused, RIL_Token t)
{
	char *cmd;
	int party;
	int err;
	party = ((int*)data)[0];

	// Make sure that party is in a valid range.
	// (Note: The Telephony middle layer imposes a range of 1 to 7.
	// It's sufficient for us to just make sure it's single digit.)
	if (party > 0 && party < 10) {
		asprintf(&cmd, "AT+CHLD=2%d", party);
		err = at_send_command(cmd, NULL);
		free(cmd);
		if(err < 0)			
			RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
		else
			RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	} else {
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	}
}

static void requestDTMF(void *data, size_t datalen __unused, RIL_Token t)
{
	char c = ((char *)data)[0];
	char *cmd;
	int err;
	asprintf(&cmd, "AT+VTS=%c", (int)c);
	err = at_send_command(cmd, NULL);
	free(cmd);
	if(err < 0)
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	else
		RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestDtmfStart(void *data, size_t datalen __unused, RIL_Token t)
{
	int err;
	char c;
	char *cmd;
	assert (datalen >= sizeof(char *));
	c = ((char *)data)[0];
	asprintf(&cmd,"AT+VTS=%c", (int)c);
	err = at_send_command(cmd,NULL);
	free(cmd);
	if(err < 0)
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	else
		 RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}

static void requestDtmfStop(RIL_Token t)
{
	int err;
	/* Send a command to cancel the DTMF tone*/
	err = at_send_command("AT",NULL);
	if(err < 0)
		RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
	else
		RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
}


/*** Callback methods from the RIL library to us ***/

/**
 * Call from RIL to us to make a RIL_REQUEST
 *
 * Must be completed with a call to RIL_onRequestComplete()
 *
 * RIL_onRequestComplete() may be called from any thread, before or after
 * this function returns.
 *
 * Will always be called from the same thread, so returning here implies
 * that the radio is ready to process another command (whether or not
 * the previous command has completed).
 */
static void onRequest (int request, void *data, size_t datalen, RIL_Token t)
{
    ATResponse *p_response;
    int err;

    RLOGD("My-onRequest: %s", requestToString(request));

    /* Ignore all requests except RIL_REQUEST_GET_SIM_STATUS
     * when RADIO_STATE_UNAVAILABLE.
     */
    RLOGD("now sState is %d",getRadioState());

    if (getRadioState() == RADIO_STATE_UNAVAILABLE
        && request != RIL_REQUEST_GET_SIM_STATUS
    ) {
        RIL_onRequestComplete(t, RIL_E_RADIO_NOT_AVAILABLE, NULL, 0);
        return;
    }

    /* Ignore all non-power requests when RADIO_STATE_OFF
     * (except RIL_REQUEST_GET_SIM_STATUS)
     */
    if (getRadioState() == RADIO_STATE_OFF
        && !(request == RIL_REQUEST_RADIO_POWER
            || request == RIL_REQUEST_GET_SIM_STATUS
            || request == RIL_REQUEST_SCREEN_STATE
            || request == RIL_REQUEST_BASEBAND_VERSION
            || request == RIL_REQUEST_GET_IMEI
            || request == RIL_REQUEST_GET_IMEISV)
    ) {
        RIL_onRequestComplete(t, RIL_E_RADIO_NOT_AVAILABLE, NULL, 0);
        return;
    }

	switch (request) {
	    case RIL_REQUEST_GET_SIM_STATUS: 
			requestGetSimStatus(t);
			break;

	    case RIL_REQUEST_GET_CURRENT_CALLS:
	        requestGetCurrentCalls(data, datalen, t);
	        break;
				
		//case RIL_REQUEST_GET_HARDWARE_CONFIG:
			//requestGetHardwareConfig(data, datalen, t);
			//break;

		//case RIL_REQUEST_GET_RADIO_CAPABILITY:
		    //requestGetRadioCapability(data, datalen, t);
		   //break;

		//case RIL_REQUEST_SET_RADIO_CAPABILITY:
		    //requestSetRadioCapability(data, datalen, t);
		    //break;

		case RIL_REQUEST_RADIO_POWER:
			requestRadioPower(data, datalen, t);
			break;

		case RIL_REQUEST_QUERY_NETWORK_SELECTION_MODE:
			requestQueryNetworkSelectionMode(data, datalen, t);
			break;

		case RIL_REQUEST_SET_NETWORK_SELECTION_MANUAL:
			requestSetNetworkSelectionManual(data, datalen, t);
			break;

		case RIL_REQUEST_SET_NETWORK_SELECTION_AUTOMATIC:
			//at_send_command("AT+COPS=0", NULL);
			break;	

		case RIL_REQUEST_QUERY_AVAILABLE_NETWORKS:
			requestQueryAvailableNetworks(data, datalen, t);
			break;

		case RIL_REQUEST_SET_PREFERRED_NETWORK_TYPE:
			requestSetPreferredNetworkType(request, data, datalen, t);
			break;
			  
		 case RIL_REQUEST_GET_PREFERRED_NETWORK_TYPE:
	        requestGetPreferredNetworkType(request, data, datalen, t);
	        break;

		case RIL_REQUEST_CDMA_SET_SUBSCRIPTION_SOURCE:
			requestCdmaSubscriptionSource(data, datalen, t);
			break;

		//case RIL_REQUEST_SET_UNSOL_CELL_INFO_LIST_RATE:
			//requestSetCellInfoListRate(data, datalen, t);
			//break;

		//case RIL_REQUEST_GET_NEIGHBORING_CELL_IDS:
			//requestNeighboringCellIds(data, datalen, t);
			//break;
		
		//case RIL_REQUEST_GET_CELL_INFO_LIST:
			//requestGetCellInfoList(data, datalen, t);
			//break;

		case RIL_REQUEST_SCREEN_STATE:
			requestScreenState(data, datalen, t);
			break;

		case RIL_REQUEST_GET_IMEI:
			requestGetIMEI(t);
			break;

		case RIL_REQUEST_GET_IMSI:
			requestGetIMSI(t);
			break;

		case RIL_REQUEST_OPERATOR:
		  	requestOperator(data, datalen, t);
		  	break;
		 
		case RIL_REQUEST_VOICE_REGISTRATION_STATE:
		case RIL_REQUEST_DATA_REGISTRATION_STATE:
		  	requestRegistrationState(request, data, datalen, t);
		  	break;

		case RIL_REQUEST_VOICE_RADIO_TECH:{
			int tech = techFromModemType(TECH(sMdmInfo));
			if (tech < 0 )
				RIL_onRequestComplete(t, RIL_E_GENERIC_FAILURE, NULL, 0);
			else
				RIL_onRequestComplete(t, RIL_E_SUCCESS, &tech, sizeof(tech));
		 }
		 	break;

		case RIL_REQUEST_SIM_IO:
			requestSIM_IO(data,datalen,t);
			break;

		case RIL_REQUEST_REPORT_STK_SERVICE_IS_RUNNING:
			requestReportStkServiceIsRunning(t);
			break;

		case RIL_REQUEST_SIGNAL_STRENGTH:
			requestSignalStrength(data, datalen, t);
			break;

		case RIL_REQUEST_DIAL:
		    requestDial(data, datalen, t);
		    break;
			
		case RIL_REQUEST_HANGUP:
		    requestHangup(data, datalen, t);
		    break;
					
	    case RIL_REQUEST_HANGUP_WAITING_OR_BACKGROUND:
	        // 3GPP 22.030 6.5.5
	        // "Releases all held calls or sets User Determined User Busy
	        //  (UDUB) for a waiting call."
			if(CALLS_NUM(sMdmInfo) > 1)
				at_send_command("AT+CHLD=0", NULL);
		    else
			    at_send_command("AT+CHUP", NULL);	
	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;
	    case RIL_REQUEST_HANGUP_FOREGROUND_RESUME_BACKGROUND:
	        // 3GPP 22.030 6.5.5
	        // "Releases all active calls (if any exist) and accepts
	        //  the other (held or waiting) call."
	        at_send_command("AT+CHLD=1", NULL);
	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;
	    case RIL_REQUEST_SWITCH_WAITING_OR_HOLDING_AND_ACTIVE:
	        // 3GPP 22.030 6.5.5
	        // "Places all active calls (if any exist) on hold and accepts
	        //  the other (held or waiting) call."
	        at_send_command("AT+CHLD=2", NULL);

	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;
	    case RIL_REQUEST_ANSWER:
	        at_send_command("ATA", NULL);

	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;
	    case RIL_REQUEST_CONFERENCE:
	        // 3GPP 22.030 6.5.5
	        // "Adds a held call to the conversation"
	        at_send_command("AT+CHLD=3", NULL);
	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;
	    case RIL_REQUEST_UDUB:
	        /* user determined user busy */
	        /* sometimes used: ATH */
	        at_send_command("ATH", NULL);
	        /* success or failure is ignored by the upper layer here.
	           it will call GET_CURRENT_CALLS and determine success that way */
	        RIL_onRequestComplete(t, RIL_E_SUCCESS, NULL, 0);
	        break;

	    case RIL_REQUEST_SEPARATE_CONNECTION:
			requestSeparateConnection(data, datalen, t);
			break;
	        
        //case RIL_REQUEST_SET_MUTE:
            //break;
        //case RIL_REQUEST_GET_MUTE:
            //break;

		case RIL_REQUEST_DTMF:
			requestDTMF(data, datalen, t);
			break;
			
		case RIL_REQUEST_DTMF_START:
			requestDtmfStart(data, datalen, t);
			break;
			
		case RIL_REQUEST_DTMF_STOP:
			requestDtmfStop(t);
			break;
			
		case RIL_REQUEST_SEND_USSD:
	     	requestSendUSSD(data, datalen, t);
	       	break;

		case RIL_REQUEST_CANCEL_USSD:
			requestCancelUSSD(t);
			break;

		case RIL_REQUEST_QUERY_CALL_WAITING:
			requestQueryCallWaiting(data, datalen, t);
			break;
			
	    case RIL_REQUEST_SET_CALL_WAITING:
	    	requestSetCallWaiting(data, datalen, t);
	    	break;
			
		case RIL_REQUEST_QUERY_CALL_FORWARD_STATUS:
			requestQueryCallForwardStatus(t);
			break;

		case RIL_REQUEST_SET_CALL_FORWARD:
			requestSetCallForward(data, t);
			break;

		case RIL_REQUEST_GET_CLIR:
			requestGetCLIR(data, datalen, t);
			break;
			
		case RIL_REQUEST_SET_CLIR:
			requestSetCLIR(data, datalen, t);
			break;

		case RIL_REQUEST_QUERY_CLIP:
			requestQueryCLIP(t);
			break;

		case RIL_REQUEST_RESET_RADIO:
			requestResetRadio(t);
			break;

		case RIL_REQUEST_SET_SUPP_SVC_NOTIFICATION:
			requestSetSuppSVCNotification(data, datalen, t);
			break;

		//case RIL_REQUEST_EXPLICIT_CALL_TRANSFER:
			//requestExplicitCallTransfer(t);
		//break;

		case RIL_REQUEST_SET_LOCATION_UPDATES:
			requestSetLocationUpdates(data, datalen, t);
			break;

		case RIL_REQUEST_LAST_CALL_FAIL_CAUSE:
			requestLastCallFailCause(t);
			break;

		case RIL_REQUEST_SETUP_DATA_CALL:
			requestSetupDataCall(data, datalen, t);
			break;

		case RIL_REQUEST_DATA_CALL_LIST:
			requestDataCallList(data, datalen, t);
			break;

		case RIL_REQUEST_DEACTIVATE_DATA_CALL:
			requestDeactivateDataCall(data, datalen, t);
			break;

		case RIL_REQUEST_LAST_DATA_CALL_FAIL_CAUSE:
			requestLastDataCallFailCause(t);
			break;

		case RIL_REQUEST_SET_INITIAL_ATTACH_APN:
			at_send_command("AT+CSTT=\"CMNET\",\"\",\"\"", NULL);
                        at_send_command("AT+CIICR",NULL);
			break;

		case RIL_REQUEST_OEM_HOOK_RAW:
			// echo back data
			RIL_onRequestComplete(t, RIL_E_SUCCESS, data, datalen);
			break;

		case RIL_REQUEST_OEM_HOOK_STRINGS:
			requestOemHookStrings(data, datalen, t);
			break;

		case RIL_REQUEST_BASEBAND_VERSION:
			requestBasebandVersion(t);
			break;


		//case RIL_REQUEST_QUERY_FACILITY_LOCK:
			//requestQueryFacilityLock(data, datalen, t);
			//break;
			
		//case RIL_REQUEST_SET_FACILITY_LOCK:
			//requestSetFacilityLock(data, datalen, t);
			//break;

		case RIL_REQUEST_DEVICE_IDENTITY:
			requestDeviceIdentity(t);
			break;

		case RIL_REQUEST_SEND_SMS:
		    requestSendSMS(data, datalen, t);
		  	break;
			case RIL_REQUEST_SEND_SMS_EXPECT_MORE:
		    requestSendSMSExpectMore(data, datalen, t);
		    break;
		case RIL_REQUEST_SMS_ACKNOWLEDGE:
		    requestSMSAcknowledge(data, datalen, t);
		    break;

		case RIL_REQUEST_WRITE_SMS_TO_SIM:
		    requestWriteSmsToSim(data, datalen, t);
		    break;
		case RIL_REQUEST_DELETE_SMS_ON_SIM: 
			requestDeleteSmsOnSim(data, datalen, t);
			break;
			case RIL_REQUEST_GET_SMSC_ADDRESS:
		    requestGetSMSCAddress(t);
		    break;
		case RIL_REQUEST_SET_SMSC_ADDRESS:
		    requestSetSMSCAddress(data, datalen, t);
		    break;
		case RIL_REQUEST_REPORT_SMS_MEMORY_STATUS:
		    requestSmsStorageStatus(data, datalen, t);
		    break;

		case RIL_REQUEST_GSM_GET_BROADCAST_SMS_CONFIG:
		    requestGSMGetBroadcastSMSConfig(t);
		    break;
		case RIL_REQUEST_GSM_SET_BROADCAST_SMS_CONFIG:
		    requestGSMSetBroadcastSMSConfig(data, datalen, t);
		    break;
		case RIL_REQUEST_GSM_SMS_BROADCAST_ACTIVATION:
		    requestGSMSMSBroadcastActivation(data, datalen, t);
		    break;

		case RIL_REQUEST_ENTER_SIM_PIN:
		case RIL_REQUEST_ENTER_SIM_PUK:
		case RIL_REQUEST_ENTER_SIM_PIN2:
		case RIL_REQUEST_ENTER_SIM_PUK2:
			requestEnterSimPin(request, data, datalen, t);
		    break;
		case RIL_REQUEST_CHANGE_SIM_PIN:
		case RIL_REQUEST_CHANGE_SIM_PIN2:
			requestChangePassword(request, data, datalen, t);
			break;
		case RIL_REQUEST_CHANGE_BARRING_PASSWORD:
			requestChangeBarringPassword(data, datalen, t);
		break;

		case RIL_REQUEST_SHUTDOWN:
		    requestShutdown(t);
		    break;
		default:
		    RLOGD("Request not supported. Tech: 0x%x",TECH(sMdmInfo));
			RLOGD("%s not supported",requestToString(request));
		    RIL_onRequestComplete(t, RIL_E_REQUEST_NOT_SUPPORTED, NULL, 0);
		    break;
	    }
}

/**
 * Call from RIL to us to find out whether a specific request code
 * is supported by this implementation.
 *
 * Return 1 for "supported" and 0 for "unsupported"
 */

static int onSupports (int requestCode __unused)
{
    //@@@ todo

    return 1;
}

static void onCancel (RIL_Token t __unused)
{
    //@@@todo

}

static const char * getVersion(void)
{
	
   	return "Cmiot ML302 Android Ril 1.0";

}

static RIL_RadioState getRadioState()
{
    return sState;
}

static void setRadioState(RIL_RadioState newState)
{
    RIL_RadioState oldState;

    pthread_mutex_lock(&s_state_mutex);

    oldState = getRadioState();

    if (s_closed > 0) {
        // If we're closed, the only reasonable state is
        // RADIO_STATE_UNAVAILABLE
        // This is here because things on the main thread
        // may attempt to change the radio state after the closed
        // event happened in another thread
        newState = RADIO_STATE_UNAVAILABLE;
    }

	RLOGD("setRadioState:oldState=%d, sState=%d, newState=%d", oldState, getRadioState(), newState);
    if (getRadioState() != newState || s_closed > 0) {
        sState = newState;

        pthread_cond_broadcast (&s_state_cond);
    }

    pthread_mutex_unlock(&s_state_mutex);

    /* do these outside of the mutex */
    if (sState != oldState) {
        RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_RADIO_STATE_CHANGED,NULL, 0);
        // Sim state can change as result of radio state change
        RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_SIM_STATUS_CHANGED,NULL, 0);

        /* FIXME onSimReady() and onRadioPowerOn() cannot be called
         * from the AT reader thread
         * Currently, this doesn't happen, but if that changes then these
         * will need to be dispatched on the request thread
         */
	if(sState == RADIO_STATE_SIM_READY || sState == RADIO_STATE_RUIM_READY || sState == RADIO_STATE_NV_READY)
	{
		onSIMReady();
		 //RIL_requestTimedCallback (onSIMReady, NULL, NULL);
	}
       else  if (sState == RADIO_STATE_SIM_NOT_READY  ||sState == RADIO_STATE_RUIM_NOT_READY  || sState == RADIO_STATE_NV_NOT_READY) {
        onRadioPowerOn();
		 //RIL_requestTimedCallback (onRadioPowerOn, NULL, NULL);
        }
    }
}


/** Returns SIM_NOT_READY on error */
static SIM_Status getSIMStatus(void)
{
    ATResponse *p_response = NULL;
    int err;
    int ret;
    char *cpinLine;
    char *cpinResult;

    RLOGD("getSIMStatus(). sState: %d",getRadioState());
    if (getRadioState() == RADIO_STATE_OFF || getRadioState() == RADIO_STATE_UNAVAILABLE) {
        ret = SIM_NOT_READY;
        goto done;
    }

    err = at_send_command_singleline("AT+CPIN?", "+CPIN:", &p_response);

    if (err < 0 || p_response->success == 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

    switch (at_get_cme_error(p_response)) {
        case CME_SUCCESS:
            break;

        case CME_SIM_NOT_INSERTED:
            ret = SIM_ABSENT;
            goto done;

        default:
            ret = SIM_NOT_READY;
            goto done;
    }

    /* CPIN? has succeeded, now look at the result */

    cpinLine = p_response->p_intermediates->line;
    err = at_tok_start (&cpinLine);

    if (err < 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

    err = at_tok_nextstr(&cpinLine, &cpinResult);

    if (err < 0) {
        ret = SIM_NOT_READY;
        goto done;
    }

   RLOGI("getSIMStatus:%s",cpinResult);	
   
    if(0 == strcmp (cpinResult, "READY"))
	ret = SIM_READY;
    else if (0 == strcmp (cpinResult, "SIM PIN")) {
        ret = SIM_PIN;
    } else if (0 == strcmp (cpinResult, "SIM PUK")) {
        ret = SIM_PUK;
    } else if (0 == strcmp (cpinResult, "SIM PIN2")) {
        ret = SIM_PIN2;
    } else if (0 == strcmp (cpinResult, "SIM PUK2"))  {
        ret = SIM_PUK2;
    }
    else{
	/* we're treating unsupported lock types as "sim absent" */
	ret = SIM_ABSENT;
    }
done:
    at_response_free(p_response);
    return ret;
}

/**
 * Fetch information about UICC card type (SIM/USIM)
 *
 * \return UICC_Type: type of UICC card.
 */
static UICC_Type getUICCType(void)
{
    ATResponse *p_response = NULL;
    char* line;
    static UICC_Type UiccType = UICC_TYPE_UNKNOWN;
    int err,type;

    if (getRadioState() == RADIO_STATE_OFF ||
        getRadioState() == RADIO_STATE_UNAVAILABLE) {
        return UICC_TYPE_UNKNOWN;
    }

    /* Only if card type is not known yet ... */
    if (UiccType == UICC_TYPE_UNKNOWN) {

        /* Get the card type */
        err = at_send_command_singleline("AT^SIMIF=1,0", "^SIMIF:", &p_response);//ml302
        if (err < 0 || p_response->success == 0)
            goto error;

        line = p_response->p_intermediates->line;
        err = at_tok_start(&line);
        if (err < 0)
            goto error;

        err = at_tok_nextint(&line, &type);
        if (err < 0)
            goto error;
	
	if(type == 0)
		UiccType = UICC_TYPE_SIM;
	else if(type == 1)
		UiccType = UICC_TYPE_USIM;
	else
		UiccType = UICC_TYPE_UNKNOWN;
    }

    RLOGI("Detected card type :%d", UiccType);

error:
    at_response_free(p_response);
    return UiccType;
}

/**
 * Get the current card status.
 *
 * This must be freed using freeCardStatus.
 * @return: On success returns RIL_E_SUCCESS
 */
static int getCardStatus(RIL_CardStatus_v6 **pp_card_status) {
    static RIL_AppStatus app_status_array[] = {
        // SIM_ABSENT = 0
        { RIL_APPTYPE_UNKNOWN, RIL_APPSTATE_UNKNOWN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_NOT_READY = 1
        { RIL_APPTYPE_SIM, RIL_APPSTATE_DETECTED, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_READY = 2
        { RIL_APPTYPE_SIM, RIL_APPSTATE_READY, RIL_PERSOSUBSTATE_READY,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // SIM_PIN = 3
        { RIL_APPTYPE_SIM, RIL_APPSTATE_PIN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN },
        // SIM_PUK = 4
        { RIL_APPTYPE_SIM, RIL_APPSTATE_PUK, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_BLOCKED, RIL_PINSTATE_UNKNOWN },
        // SIM_NETWORK_PERSONALIZATION = 5
        { RIL_APPTYPE_SIM, RIL_APPSTATE_SUBSCRIPTION_PERSO, RIL_PERSOSUBSTATE_SIM_NETWORK,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN },
        // RUIM_ABSENT = 6
        { RIL_APPTYPE_UNKNOWN, RIL_APPSTATE_UNKNOWN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // RUIM_NOT_READY = 7
        { RIL_APPTYPE_RUIM, RIL_APPSTATE_DETECTED, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // RUIM_READY = 8
        { RIL_APPTYPE_RUIM, RIL_APPSTATE_READY, RIL_PERSOSUBSTATE_READY,
          NULL, NULL, 0, RIL_PINSTATE_UNKNOWN, RIL_PINSTATE_UNKNOWN },
        // RUIM_PIN = 9
        { RIL_APPTYPE_RUIM, RIL_APPSTATE_PIN, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN },
        // RUIM_PUK = 10
        { RIL_APPTYPE_RUIM, RIL_APPSTATE_PUK, RIL_PERSOSUBSTATE_UNKNOWN,
          NULL, NULL, 0, RIL_PINSTATE_ENABLED_BLOCKED, RIL_PINSTATE_UNKNOWN },
        // RUIM_NETWORK_PERSONALIZATION = 11
        { RIL_APPTYPE_RUIM, RIL_APPSTATE_SUBSCRIPTION_PERSO, RIL_PERSOSUBSTATE_SIM_NETWORK,
           NULL, NULL, 0, RIL_PINSTATE_ENABLED_NOT_VERIFIED, RIL_PINSTATE_UNKNOWN }
    };
    RIL_CardState card_state;
    int num_apps;

    int sim_status = getSIMStatus();
    if (sim_status == SIM_ABSENT) {
        card_state = RIL_CARDSTATE_ABSENT;
        num_apps = 0;
    } else {
        card_state = RIL_CARDSTATE_PRESENT;
        num_apps = 1;
    }

    // Allocate and initialize base card status.
    RIL_CardStatus_v6 *p_card_status = malloc(sizeof(RIL_CardStatus_v6));
    p_card_status->card_state = card_state;
    p_card_status->universal_pin_state = RIL_PINSTATE_UNKNOWN;
    p_card_status->gsm_umts_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->cdma_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->ims_subscription_app_index = RIL_CARD_MAX_APPS;
    p_card_status->num_applications = num_apps;

    // Initialize application status
    int i;
    for (i = 0; i < RIL_CARD_MAX_APPS; i++) {
        p_card_status->applications[i] = app_status_array[SIM_ABSENT];
    }

    // Pickup the appropriate application status
    // that reflects sim_status for gsm.
    if (num_apps != 0) {
        UICC_Type uicc_type = getUICCType();
        // Only support one app, gsm
        p_card_status->num_applications = 1;
        p_card_status->gsm_umts_subscription_app_index = 0;
        //p_card_status->cdma_subscription_app_index = 1;

        // Get the correct app status
        p_card_status->applications[0] = app_status_array[sim_status];
        //p_card_status->applications[1] = app_status_array[sim_status + RUIM_ABSENT];
		if (uicc_type == UICC_TYPE_SIM){
        	RLOGI("[Card type discovery]: Legacy SIM");
	    	p_card_status->applications[0].app_type = RIL_APPTYPE_SIM;
		}
        else { /* defaulting to USIM */
            RLOGI("[Card type discovery]: USIM");
            p_card_status->applications[0].app_type = RIL_APPTYPE_USIM;
		}
    }

    *pp_card_status = p_card_status;
    return RIL_E_SUCCESS;
}

/**
 * Free the card status returned by getCardStatus
 */
static void freeCardStatus(RIL_CardStatus_v6 *p_card_status) 
{
    free(p_card_status);
}

/**
 * SIM ready means any commands that access the SIM will work, including:
 *  AT+CPIN, AT+CSMS, AT+CNMI, AT+CRSM
 *  (all SMS-related commands)
 */

static void pollSIMState (void *param __unused)
{
    ATResponse *p_response;
    int ret;

    if ((getRadioState() != RADIO_STATE_SIM_NOT_READY)
        && (getRadioState() != RADIO_STATE_RUIM_NOT_READY))
    {
        return;
    }

    switch(getSIMStatus()) {
	case SIM_READY:
            RLOGI("pollSIMState:SIM_READY");
	    setRadioState(RADIO_STATE_SIM_READY);	
            //RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_SIM_STATUS_CHANGED, NULL, 0);
       	    return;

        case SIM_NOT_READY:
            RIL_requestTimedCallback (pollSIMState, NULL, &TIMEVAL_SIMPOLL);
            return;

        case SIM_ABSENT:
        case SIM_PIN:
        case SIM_PUK:
        case SIM_NETWORK_PERSONALIZATION:
        default:
            RLOGI("SIM ABSENT or LOCKED");
	    setRadioState(RADIO_STATE_SIM_LOCKED_OR_ABSENT);
           // RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_SIM_STATUS_CHANGED, NULL, 0);
        return;
    }
}

/** returns 1 if on, 0 if off, and -1 on error */
static int isRadioOn()
{
    ATResponse *p_response = NULL;
    int err;
    char *line;
    char ret;

    err = at_send_command_singleline("AT+CFUN?", "+CFUN:", &p_response);

    if (err < 0 || p_response->success == 0) {
        // assume radio is off
        goto error;
    }

    line = p_response->p_intermediates->line;

    err = at_tok_start(&line);
    if (err < 0) goto error;

    err = at_tok_nextbool(&line, &ret);
    if (err < 0) goto error;

    at_response_free(p_response);

    return (int)ret;

error:

    at_response_free(p_response);
    return -1;
}

static void getIccid(void)
{
	ATResponse *p_response = NULL;
	char *line = NULL;
	int err = 0;

	err = at_send_command_singleline("AT+CCID", "+CCID:", &p_response);

	if (err < 0 || p_response->success == 0) {
	   property_set("ril.cmiot.iccid", "");
	} else {
		line = p_response->p_intermediates->line;
		err = at_tok_start(&line);
		property_set("ril.cmiot.iccid", line);
	}
	at_response_free(p_response);
}

static int switchToinMode(void)
{
	ATResponse *p_response = NULL;
	char *line;
	int err,ret = -1;
	int mode = 0;
	
    at_handshake();

	do{
		/* Check Modem capabilities */
		err = at_send_command_singleline("AT+ZLANENABLE?", "+ZLANENABLE:", &p_response);
		if (err < 0 || p_response->success == 0)
			break;
		line = p_response->p_intermediates->line;
		err = at_tok_start(&line);
		if (err < 0) break;

		err = at_tok_nextint(&line, &mode);
		break;
	}while(1);
	
	
	if(mode == 1)
	{
		//switch ZLANENABLE 1 --> 0
		at_send_command("AT+ZLANENABLE=0",NULL);   //reboot
		ret = 0;
	}
	
    at_response_free(p_response);
	return ret;
}

/* find ECM Port,link:eth0 usb0

	0,success,-1,error
*/
static int GetECMPort()
{
    char* temp = NULL;
    DIR*    pDir_usb = NULL;                         //  /sys/class/net 
    FILE*   pfmodalias = NULL;                      //  /sys/class/net/%s/modalias 
    char    cdirmodalias[1024] = "";      
    char    cmodalias[1024] = "";         
    int     ret = -1;
    memset(cdirmodalias, '\0', sizeof(cdirmodalias));
    memset(cmodalias, '\0', sizeof(cmodalias));
    struct dirent* pNext_usb = NULL;

    pDir_usb = opendir("/sys/class/net");
    if (NULL == pDir_usb)
    {
        return ret;
    }
    while (NULL != (pNext_usb = readdir(pDir_usb)))
    {
        snprintf(cdirmodalias, sizeof(cdirmodalias) - 1, "/sys/class/net/%s/device/modalias", pNext_usb->d_name);

        if (NULL != (pfmodalias = fopen(cdirmodalias, "r")))
        {
            temp = fgets(cmodalias, 1024, pfmodalias);
            if (NULL == temp)
            {
                RLOGD("GetECMPort gets FAIL!!");
            }

	    if (NULL != strstr(cmodalias, "v1782") || NULL != strstr(cmodalias, "v1782"))
            {
                snprintf(PORT_ECM_KEYNAME, sizeof(PORT_ECM_KEYNAME) - 1, "%s", pNext_usb->d_name); //coverity=SECURE_CODING
                RLOGI("ECM port is %s\n", PORT_ECM_KEYNAME);
                ret = 0;
            }
            fclose(pfmodalias);
        }
        /*End modify by d00141585 for DTS2013091610549 2013-09-20*/
    }
    if (NULL != pDir_usb)
    {
        closedir(pDir_usb);
    }
    return ret;
}

/**
 * Find out if our modem is GSM, CDMA or both (Multimode)
 */
static void probeForModemMode(ModemInfo *info)
{
	int err;
	assert (info);
    // TODO: find out if modem really supports WCDMA/LTE
    info->supportedTechs = MDM_GSM | MDM_WCDMA | MDM_LTE;
    info->currentTech = MDM_LTE;
    info->usePPP = pppSupported();
	info->useVoLTE = 0;
    info->isPppStart = PPP_CLOSED;
	info->callsNum = 0;
    RLOGI("Found GSM Modem");

	if(info->usePPP == 0){
       	err = GetECMPort();
		if(err < 0){		
			snprintf(PORT_ECM_KEYNAME, sizeof(PORT_ECM_KEYNAME) - 1, "%s", "ecm get error");
			info->usePPP = 1;
		}
	}else{		
		snprintf(PORT_ECM_KEYNAME, sizeof(PORT_ECM_KEYNAME) - 1, "%s", "ppp0");
	}
	memset(&rmnetParam,0,sizeof(rmnetParam));
	
	if(USE_PPP(sMdmInfo)){
		at_send_command("AT+ZNETCARDTYPE=3",NULL);
	}else{
		at_send_command("AT+ZNETCARDTYPE=1",NULL);
}
}

static void initMoreFromProperty(void)
{
	char cmvol[PROPERTY_VALUE_MAX] = {'\0'};
	int err;
	char *cmd;
	
	err = property_get("ril.cmiot.cmvol", cmvol, NULL);

	if((err >= 0) && (strcmp(cmvol, "") != 0)){ 
		asprintf(&cmd,"AT+CMVOL=%s",cmvol);
		err = at_send_command(cmd,NULL);
		if(err < 0)
			RLOGE("set cmvol error!");

		free(cmd);
	}
	else{
		RLOGD("can't find property:ril.cmiot.cmvol");
	}
}

/**
 * Initialize everything that can be configured while we're still in
 * AT+CFUN=0
 */
static void initializeCallback(void *param __unused)
{
    //ATResponse *p_response = NULL;
    int err;

    RLOGI("add by tim initializeCallback()");
    setRadioState (RADIO_STATE_OFF);

    at_handshake();
	//at_send_command("AT", NULL);
	
    probeForModemMode(sMdmInfo);

    /* note: we don't check errors here. Everything important will
       be handled in onATTimeout and onATReaderClosed */

    /*  atchannel is tolerant of echo but it must */
    /*  have verbose result codes */
    at_send_command("ATE0Q0V1", NULL);

    /*  No auto-answer */
    at_send_command("ATS0=0", NULL);

    /*  Extended errors */
    at_send_command("AT+CMEE=1", NULL);
	
	/*get iccid*/
	getIccid();
    
    /* assume radio is off on error */
    if (isRadioOn() > 0) {
        setRadioState (RADIO_STATE_SIM_NOT_READY);
    }
}

static void waitForClose()
{
	RLOGI("entering waitForClose");

	pthread_mutex_lock(&s_state_mutex);

    	while (s_closed == 0) {
        	pthread_cond_wait(&s_state_cond, &s_state_mutex);
    	}

    	pthread_mutex_unlock(&s_state_mutex);

	RLOGI("leaving waitForClose");
}

#define RIL_PS_NET_NOTICE_INVALID 	-1
#define RIL_PS_NET_NOTICE_DEACT 	0
#define RIL_PS_NET_NOTICE_ACT 		1
#define RIL_PS_NET_NOTICE_MODIFY 	2
#define RIL_PS_NET_NOTICE_REACT 	3


static void unsolicitedCGEV(const char * s)
{
	char *line = NULL;
	int action = RIL_PS_NET_NOTICE_INVALID;
	line = strdup(s);
	
	if(NULL != strstr(line, "DEACT")){
		action = RIL_PS_NET_NOTICE_DEACT;		
	}else if(NULL != strstr(line, "ACT")){
		action = RIL_PS_NET_NOTICE_ACT;
	}else if(NULL != strstr(line, "MODIFY")){
		action = RIL_PS_NET_NOTICE_MODIFY;
	}else if(NULL != strstr(line, "REACT")){
		action = RIL_PS_NET_NOTICE_REACT;
	}else{
		action = RIL_PS_NET_NOTICE_INVALID;
	}

	RLOGD("%s:action:%d", __FUNCTION__, action);

	if(action == RIL_PS_NET_NOTICE_DEACT)
		onDeactiveDataCallList();
	else
		RIL_requestTimedCallback (onDataCallListChanged, NULL, NULL);

	free(line);
}

static void unsolicitedMode(const char * s)
{
	char *line = NULL;
	char *linesave = NULL;
	int err;
	int value,tech;

	line = strdup(s);
	linesave = line;
	err = at_tok_start(&line);
	if (err < 0)
		goto error;

	err = at_tok_nextint(&line, &value);
	if (err < 0)
		goto error;
	
	switch(value)
	{
		case 3:
			tech = MDM_GSM;
			break;
		case 5:
			tech = MDM_WCDMA;
			break;
		case 15:
			tech = MDM_TD_SCDMA;
			break;
		case 17:
			tech = MDM_LTE;
			break;
		default:
			//if(PPPD_START(sMdmInfo) == PPP_STARTED)
			//onDeactiveDataCallList();
			//RIL_onUnsolicitedResponse(RIL_UNSOL_RESPONSE_VOICE_NETWORK_STATE_CHANGED, NULL, 0);
			goto error;
	}

	int oldTech = TECH(sMdmInfo);
	RLOGD("unsolicitedMode Tech change (%d => %d)", oldTech, tech);
	if (tech != oldTech) {
		TECH(sMdmInfo) = tech;
		if (techFromModemType(tech) != techFromModemType(oldTech)) 
		{
			int tech = techFromModemType(TECH(sMdmInfo));
			if (tech > 0 ) {
				RIL_onUnsolicitedResponse(RIL_UNSOL_VOICE_RADIO_TECH_CHANGED,&tech, sizeof(tech));
				}
			}
	}
	
	free(linesave);
	return;

error:
	free(linesave);
	RLOGE("unsolicitedMode: Error parameter in ind msg: %s", s);
	return;
}

static void unsolicitedNitzTime(const char * s)
{
	char *response;
	char *line = NULL;
	char *linesave = NULL;
	int err;
	char *univer_time,*time_zone;
	int dst;
	char *skip;
	
    /* Higher layers expect a NITZ string in this format:
     *  08/10/28,19:08:37-20,1 (yy/mm/dd,hh:mm:ss(+/-)tz,dst)
     */
	//+ZMMI: +32,0,"","","18/07/05,02:26:59",""
	
	line = strdup(s);
	linesave = line;
	err = at_tok_start(&line);
		if(err < 0) goto error;
	err = at_tok_nextstr(&line, &time_zone);
		if(err < 0) goto error;
	err = at_tok_nextint(&line, &dst);
		if(err < 0) goto error;
	err = at_tok_nextstr(&line, &skip);
		if(err < 0) goto error;
	err = at_tok_nextstr(&line, &skip);
		if(err < 0) goto error;
	err = at_tok_nextstr(&line, &univer_time);
		if(err < 0) goto error;

	asprintf(&response,"%s%s,%d",univer_time,time_zone,dst);
	RLOGE("NITZ:%s\n", response);
	RIL_onUnsolicitedResponse (RIL_UNSOL_NITZ_TIME_RECEIVED,response, strlen(response));
	free(response);
	free(linesave);
	return;
	
error:
	free(response);
	free(linesave);
	RLOGE("unsolicitedNitzTime: Error parameter in ind msg: %s", s);
	return;

}

static void unsolicitedZGIPDNS(const char * s)
{
	char *line = NULL,*p;
	char *linesave = NULL;
	int err;
	char *type,*local_ip,*gw,*dns1,*dns2;
	int cid_num,cid,commas;

	//+ZGIPDNS: 1,1,"IP","10.222.60.102","0.0.0.0","112.5.230.54","211.143.147.120"
		
	line = strdup(s);
	linesave = line;
	
	err = at_tok_start(&line);
	if (err < 0)
		goto error; 

	commas = 0;
    for (p = line ; *p != '\0' ;p++) {
        if (*p == ',') commas++;
    }

	if(commas < 6)
		goto error;
	
	err =  at_tok_nextint(&line,&cid_num);
	if (err < 0)
		goto error; 

	//if(cid_num >1)

	err =  at_tok_nextint(&line,&cid);
	if (err < 0)
		goto error; 

	err = at_tok_nextstr(&line,&type);
	if (err < 0)
		goto error; 

	err = at_tok_nextstr(&line,&local_ip);
	if (err < 0)
		goto error; 

	err = at_tok_nextstr(&line,&gw);
	if (err < 0)
		goto error; 

	err = at_tok_nextstr(&line,&dns1);
	if (err < 0)
		goto error; 

	err = at_tok_nextstr(&line,&dns2);
	if (err < 0)
		goto error; 

	RLOGD("unsolicitedZGIPDNS:local_ip=%s,dns1=%s,dns2=%s,gw=%s",local_ip,dns1,dns2,gw);

	property_set("net.ppp0.local-ip",local_ip);
	property_set("net.ppp0.dns1",dns1);
	property_set("net.ppp0.dns2",dns2);
	property_set("net.ppp0.remote-ip",gw);
	
	free(linesave);
	return;

error:
	free(linesave);
	RLOGE("unsolicitedZGIPDNS: Error parameter in ind msg: %s", s);
	return;
}

static void unsolicitedIMSVMODE(const char * s)
{
	char *line = NULL,*p;
	char *linesave = NULL;
	int err,imsvmode;
	
	line = strdup(s);
	linesave = line;
	
	err = at_tok_start(&line);
	if (err < 0)
		goto error; 
	
	err = at_tok_nextint(&line,&imsvmode);
	if (err < 0)
		goto error; 

	USE_VOLTE(sMdmInfo) = imsvmode;
	
	free(linesave);
	return;

error:
	free(linesave);
	RLOGE("unsolicitedIMSVMODE: Error parameter in ind msg: %s", s);
	return;
}

/**
 * Called by atchannel when an unsolicited line appears
 * This is called on atchannel's reader thread. AT commands may
 * not be issued here
 */
static void onUnsolicited (const char *s, const char *sms_pdu)
{
    char *line = NULL, *p;
    int err;

    /* Ignore unsolicited responses until we're initialized.
     * This is OK because the RIL library will poll for initial state
     */

   //return; //add by test	
   	
    if (getRadioState() == RADIO_STATE_UNAVAILABLE) {
        return;
    } 

    if (strStartsWith(s, "+ZMMI:")) {
		unsolicitedNitzTime(s);
    } else if (strStartsWith(s,"+CRING:")
                || strStartsWith(s,"RING")
                || strStartsWith(s,"NO CARRIER")
                || strStartsWith(s,"+CCWA:")
				//|| strStartsWith(s,"+ZCWA")
                || strStartsWith(s,"+CLIP:")
				|| strStartsWith(s,"^ORIG:")
				|| strStartsWith(s,"^DSCI:")
				|| strStartsWith(s,"^CONF:")
				|| strStartsWith(s,"^CONN:")
				|| strStartsWith(s,"^CEND:")
				|| strStartsWith(s,"+ZCPI:")

    ) {
        RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_CALL_STATE_CHANGED,NULL, 0);
    } else if (strStartsWith(s,"+CREG:")
   	        ||strStartsWith(s,"+CGREG:")
                || strStartsWith(s,"+CEREG:")
    ) {
        RIL_onUnsolicitedResponse (RIL_UNSOL_RESPONSE_VOICE_NETWORK_STATE_CHANGED,NULL, 0);
    }else if (strStartsWith(s, "+CMTI:")
			|| strStartsWith(s, "+CDSI:")) {
		newSmsOrStatusReportOnSIM(s);
	}else if (strStartsWith(s, "+CMT:")) {
		newSmsReceived(sms_pdu);
    }else if (strStartsWith(s, "+CDS:")) {
    	newSmsStatusReportReceived(sms_pdu);
    }else if(strStartsWith(s, "+CBM:")){ 
		newBroadcastSms(sms_pdu);
	}else if(strStartsWith(s, "+ZMGSF:")){
		simSmsStorageFull(s,sms_pdu);
	}else if (strStartsWith(s, "+CGEV:")) {
        /* Really, we can ignore NW CLASS and ME CLASS events here,
         * but right now we don't since extranous
         * RIL_UNSOL_DATA_CALL_LIST_CHANGED calls are tolerated
         */
        /* can't issue AT commands here -- call on main thread */
        unsolicitedCGEV(s);
    }else if(strStartsWith(s, "^MODE:")){
		unsolicitedMode(s);
    }else if(strStartsWith(s, "+ZGIPDNS:")){
		//unsolicitedZGIPDNS(s);
    }else if(strStartsWith(s, "+ZLTENOCELL")){
		//if(PPPD_START(sMdmInfo) == PPP_STARTED)
    	//onDeactiveDataCallList();
	}else if(strStartsWith(s, "+IMSVMODE:")){
	
	}else if (strStartsWith(s, "+CFUN: 0")) {
        setRadioState(RADIO_STATE_OFF);
    }
}

/* Called on command or reader thread */
static void onATReaderClosed()
{
    RLOGI("AT channel closed\n");
    at_close();
    s_closed = 1;

    setRadioState (RADIO_STATE_UNAVAILABLE);
}

/* Called on command thread */
static void onATTimeout()
{
    RLOGI("AT channel timeout; closing\n");
    at_close();

    s_closed = 1;

    /* FIXME cause a radio reset here */

    setRadioState (RADIO_STATE_UNAVAILABLE);
}

static void usage(char *s)
{
#ifdef RIL_SHLIB
    fprintf(stderr, "reference-ril requires: -p <tcp port> or -d /dev/tty_device\n");
#else
    fprintf(stderr, "usage: %s [-p <tcp port>] [-d /dev/tty_device]\n", s);
    exit(-1);
#endif
}

static void *pppMainLoop()
{	
	int err;
	int retry;
	sleep(10);
	for(retry = 0;retry < 3;retry++){
		if(USE_PPP(sMdmInfo) == 1)
			break;
		
		sleep(3);
	}

	RLOGD("pppMainLoop:retry=%d",retry);
	if(retry >= 3 && USE_PPP(sMdmInfo) == 0)
		return ((void *)0);

	RLOGD("pppMainLoop starting!\n");
	for(;;){
		if(s_closed == 1)
			break;
		if(PPPD_START(sMdmInfo) == PPP_STARTED){
			err = checkPPPIsUPByPPPLoop(2);
			if((err < 0) && (PPPD_START(sMdmInfo) == PPP_STARTED)){
				RLOGD("pppMainLoop:can not detect PPP state!\n");
				onDeactiveDataCallList();	
			}
		}
		sleep(3);
   }

	return ((void *)0);
}

static int startPppCheckPthread()
{	
	int ret = -1;
	pthread_attr_t attr;
	pthread_attr_init (&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&s_tid_mainloop, &attr, pppMainLoop,  NULL);
	return ret;
}

static void *mainLoop(void *param __unused)
{
    int fd;
    int ret;

   	RLOGI("mainLoop() starting!");
   
    AT_DUMP("== ", "entering mainLoop()", -1 );
    at_set_on_reader_closed(onATReaderClosed);
    at_set_on_timeout(onATTimeout);

    for (;;) {
        fd = -1;
        while  (fd < 0) {
	    if (s_device_path != NULL) {
                fd = open (s_device_path, O_RDWR | O_NOCTTY);
                if ( fd >= 0 && !memcmp( s_device_path, "/dev/ttyUSB", 11 ) ) {
                    /* disable echo on serial ports */
                    struct termios  ios;
                    tcgetattr( fd, &ios );
                    ios.c_lflag = 0;  /* disable ECHO, ICANON, etc... */
                    tcsetattr( fd, TCSANOW, &ios );
                }
            }

            if (fd < 0) {
                perror ("opening AT interface. retrying...");
                sleep(10);
                /* never returns */
            }
        }

        s_closed = 0;
        ret = at_open(fd, onUnsolicited);

        if (ret < 0) {
            RLOGE ("AT error %d on at_open\n", ret);
            return 0;
        }
		sleep(5);

		//property_set("ril.cmiot.cmvol", "10");   //for test

		#ifdef CMIOT_DAIL_ECM   // ECM \B2\A6\BAš\A3Ĭ\C8\CFΪPPP\B2\A6\BA\C5
		
		property_set("ril.cmiot.dialmode","ecm"); 
		
		#endif
		if((ret = startPppCheckPthread()) < 0){
	    	RLOGE("%s: start_ppp_check_pthread() failed: %s", __FUNCTION__, strerror(errno));
			return 0;
        }
		
        RIL_requestTimedCallback(initializeCallback, NULL, &TIMEVAL_0);

        // Give initializeCallback a chance to dispatched, since
        // we don't presently have a cancellation mechanism
        sleep(1);

        waitForClose();
        RLOGI("Re-opening after close");
    }
}

#ifdef RIL_SHLIB
const RIL_RadioFunctions *RIL_Init(const struct RIL_Env *env, int argc, char **argv)
{
    int ret;
    int fd = -1;
    int opt;
    pthread_attr_t attr;

    s_rilenv = env;

    while ( -1 != (opt = getopt(argc, argv, "p:d:s:c:"))) {
        switch (opt) {
            case 'p':
                s_port = atoi(optarg);
                if (s_port == 0) {
                    usage(argv[0]);
                    return NULL;
                }
                RLOGI("Opening loopback port %d\n", s_port);
            break;

            case 'd':
                s_device_path = optarg;
                RLOGI("Opening tty device %s\n", s_device_path);
            break;

            case 's':
                s_device_path   = optarg;
                s_device_socket = 1;
                RLOGI("Opening socket %s\n", s_device_path);
            break;

            case 'c':
                RLOGI("Client id received %s\n", optarg);
            break;

            default:
                usage(argv[0]);
                return NULL;
        }
    }

    if (s_port < 0 && s_device_path == NULL) {
        usage(argv[0]);
        return NULL;
    }

    sMdmInfo = calloc(1, sizeof(ModemInfo));
    if (!sMdmInfo) {
        RLOGE("Unable to alloc memory for ModemInfo");
        return NULL;
    }

    pthread_attr_init (&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    ret = pthread_create(&s_tid_mainloop, &attr, mainLoop, NULL);
	
    return &s_callbacks;
}
#else /* RIL_SHLIB */
int main (int argc, char **argv)
{
    int ret;
    int fd = -1;
    int opt;

    while ( -1 != (opt = getopt(argc, argv, "p:d:"))) {
        switch (opt) {
            case 'p':
                s_port = atoi(optarg);
                if (s_port == 0) {
                    usage(argv[0]);
                }
                RLOGI("Opening loopback port %d\n", s_port);
            break;

            case 'd':
                s_device_path = optarg;
                RLOGI("Opening tty device %s\n", s_device_path);
            break;

            case 's':
                s_device_path   = optarg;
                s_device_socket = 1;
                RLOGI("Opening socket %s\n", s_device_path);
            break;

            default:
                usage(argv[0]);
        }
    }

    if (s_port < 0 && s_device_path == NULL) {
        usage(argv[0]);
    }

    RIL_register(&s_callbacks);

    mainLoop(NULL);

    return 0;
}

#endif /* RIL_SHLIB */
